# Copyright 2011 OpenStack Foundation
# Copyright 2011, Piston Cloud Computing, Inc.
# Copyright 2011 Nebula, Inc.
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import copy
import typing as ty

import typing_extensions as ty_ext

from keystoneauth1 import discover
from keystoneauth1 import exceptions

_SERVICE_CATALOG_T = list[dict[str, ty.Any]]


class ServiceCatalog(metaclass=abc.ABCMeta):
    """Helper methods for dealing with a Keystone Service Catalog."""

    def __init__(self, catalog: _SERVICE_CATALOG_T):
        self._catalog = catalog

    @classmethod
    @abc.abstractmethod
    def from_token(cls, token: dict[str, ty.Any]) -> ty_ext.Self:
        """Retrieve the service catalog from a token.

        :param token:
        :returns: A service catalog.
        """

    def _get_endpoint_region(self, endpoint: dict[str, str]) -> str | None:
        return endpoint.get('region_id') or endpoint.get('region')

    @property
    def catalog(self) -> _SERVICE_CATALOG_T:
        """Return the raw service catalog content, mostly useful for debugging.

        Applications should avoid this and use accessor methods instead.
        However, there are times when inspecting the raw catalog can be useful
        for analysis and other reasons.
        """
        return self._catalog

    @abc.abstractmethod
    def is_interface_match(
        self, endpoint: dict[str, str], interface: str
    ) -> bool:
        """Helper function to normalize endpoint matching across v2 and v3.

        :returns: True if the provided endpoint matches the required
                  interface otherwise False.
        """

    @staticmethod
    def normalize_interface(interface: str) -> str:
        """Handle differences in the way v2 and v3 catalogs specify endpoint.

        Both v2 and v3 must be able to handle the endpoint style of the other.
        For example v2 must be able to handle a 'public' interface and
        v3 must be able to handle a 'publicURL' interface.

        :returns: the endpoint string in the format appropriate for this
                  service catalog.
        """
        return interface

    def _normalize_endpoints(
        self, endpoints: list[dict[str, ty.Any]]
    ) -> list[dict[str, ty.Any]]:
        """Translate endpoint description dicts into v3 form.

        Takes a raw endpoint description from the catalog and changes
        it to be in v3 format. It also saves a copy of the data in
        raw_endpoint so that it can be returned by methods that expect the
        actual original data.

        :param list endpoints: List of endpoint description dicts

        :returns: List of endpoint description dicts in v3 format
        """
        new_endpoints = []
        for endpoint in endpoints:
            raw_endpoint = endpoint.copy()
            new_endpoint = endpoint.copy()
            new_endpoint['raw_endpoint'] = raw_endpoint
            new_endpoints.append(new_endpoint)
        return new_endpoints

    def _denormalize_endpoints(
        self, endpoints: list[discover.EndpointData]
    ) -> list[dict[str, ty.Any] | None]:
        """Return original endpoint description dicts.

        Takes a list of EndpointData objects and returns the original
        dict that was returned from the catalog.

        :param list endpoints: List of `keystoneauth1.discover.EndpointData`

        :returns: List of endpoint description dicts in original catalog format
        """
        return [endpoint.raw_endpoint for endpoint in endpoints]

    def normalize_catalog(self) -> list[dict[str, ty.Any]]:
        """Return the catalog normalized into v3 format."""
        catalog = []
        for service in copy.deepcopy(self._catalog):
            if 'type' not in service:
                continue

            # NOTE(jamielennox): service_name is different. It is not available
            # in API < v3.3. If it is in the catalog then we enforce it, if it
            # is not then we don't because the name could be correct we just
            # don't have that information to check against. Set to None so
            # that checks will naturally work.
            service.setdefault('name', None)

            # NOTE(jamielennox): there is no such thing as a service_id in v2
            # similarly to service_name.
            service.setdefault('id', None)

            service['endpoints'] = self._normalize_endpoints(
                service.get('endpoints', [])
            )

            for endpoint in service['endpoints']:
                endpoint['region_name'] = self._get_endpoint_region(endpoint)
                endpoint.setdefault('id', None)
            catalog.append(service)
        return catalog

    def _get_interface_list(
        self, interface: str | list[str] | tuple[str] | set[str] | None
    ) -> list[str]:
        if not interface:
            return []
        if not isinstance(interface, list | tuple | set):
            interface = [interface]
        return [self.normalize_interface(i) for i in interface]

    def get_endpoints_data(
        self,
        service_type: str | None = None,
        interface: str | list[str] | tuple[str] | set[str] | None = None,
        region_name: str | None = None,
        service_name: str | None = None,
        service_id: str | None = None,
        endpoint_id: str | None = None,
    ) -> dict[str, list[discover.EndpointData]]:
        """Fetch and filter endpoint data for the specified service(s).

        Returns endpoints for the specified service (or all) containing
        the specified type (or all) and region (or all) and service name.

        If there is no name in the service catalog the service_name check will
        be skipped.  This allows compatibility with services that existed
        before the name was available in the catalog.

        Valid interface types: `public` or `publicURL`,
                               `internal` or `internalURL`,
                               `admin` or 'adminURL`

        :param string service_type: Service type of the endpoint.
        :param interface: Type of endpoint. Can be a single value or a list
                          of values. If it's a list of values, they will be
                          looked for in order of preference.
        :param string region_name: Region of the endpoint.
        :param string service_name: The assigned name of the service.
        :param string service_id: The identifier of a service.
        :param string endpoint_id: The identifier of an endpoint.

        :returns: a dict, keyed by service_type, of lists of EndpointData
        """
        interfaces = self._get_interface_list(interface)

        matching_endpoints: dict[str, list[discover.EndpointData]] = {}

        for service in self.normalize_catalog():
            if service_type and not discover._SERVICE_TYPES.is_match(
                service_type, service['type']
            ):
                continue

            if (
                service_name
                and service['name']
                and service_name != service['name']
            ):
                continue

            if service_id and service['id'] and service_id != service['id']:
                continue

            matching_endpoints.setdefault(service['type'], [])

            for endpoint in service.get('endpoints', []):
                if interfaces and endpoint['interface'] not in interfaces:
                    continue
                if region_name and region_name != endpoint['region_name']:
                    continue
                if endpoint_id and endpoint_id != endpoint['id']:
                    continue
                if not endpoint['url']:
                    continue

                matching_endpoints[service['type']].append(
                    discover.EndpointData(
                        catalog_url=endpoint['url'],
                        service_type=service['type'],
                        service_name=service['name'],
                        service_id=service['id'],
                        interface=endpoint['interface'],
                        region_name=endpoint['region_name'],
                        endpoint_id=endpoint['id'],
                        raw_endpoint=endpoint['raw_endpoint'],
                    )
                )

        if not interfaces:
            return self._endpoints_by_type(service_type, matching_endpoints)

        ret: dict[str, list[discover.EndpointData]] = {}
        for matched_service_type, endpoints in matching_endpoints.items():
            if not endpoints:
                ret[matched_service_type] = []
                continue

            matches_by_interface: dict[
                str | None, list[discover.EndpointData]
            ] = {}
            for endpoint in endpoints:
                if endpoint.interface is None:
                    # interface will always be set in a real deployment since
                    # it is non-nullable: if we get here, we're in a test or
                    # broken
                    continue

                matches_by_interface.setdefault(endpoint.interface, [])
                matches_by_interface[endpoint.interface].append(endpoint)

            best_interface = [
                i for i in interfaces if i in matches_by_interface.keys()
            ][0]
            ret[matched_service_type] = matches_by_interface[best_interface]

        return self._endpoints_by_type(service_type, ret)

    def _endpoints_by_type(
        self,
        requested: str | None,
        endpoints: dict[str, list[discover.EndpointData]],
    ) -> dict[str, list[discover.EndpointData]]:
        """Get the approrpriate endpoints from the list of given endpoints.

        Per the service type alias rules:

        If a user requests a service by its proper name and that matches, win.

        If a user requests a service by its proper name and only a single alias
        matches, win.

        If a user requests a service by its proper name and more than one alias
        matches, choose the first alias from the list given.

        Do the "first alias" match after the other filters, as they might limit
        the number of choices for us otherwise.

        :param str requested:
            The service_type as requested by the user.
        :param dict endpoints:
            A dictionary keyed by found service_type. Values are opaque to
            this method.

        :returns:
            Dict of service_type/endpoints filtered for the appropriate
            service_type based on alias matching rules.
        """
        if not requested or not discover._SERVICE_TYPES.is_known(requested):
            # The user did not request a service we have any alias information
            # about, or did not request a service, which means that we cannot
            # further filter the list.
            return endpoints

        if len(endpoints) < 2:
            # There is at most one type found from the initial pass through
            # the catalog. Nothing further to do.
            return endpoints

        # At this point, the user has requested a type, we do know things
        # about aliases for that type, and we've found more than one match.
        # We must filter out additional types, otherwise clouds that register
        # the same endpoint twice as part of a migration will confuse users.

        # Only return the one the user requested if there's an exact match
        # and there is data for it. There might not be data for this match
        # if there are other filters that excluded it from consideration
        # after we accepted it as an alias.
        if endpoints.get(requested):
            return {requested: endpoints[requested]}

        # We've matched something that isn't exactly what the user requested.
        # Look at the possible types for this service in order or priority and
        # return the first match.
        for alias in discover._SERVICE_TYPES.get_all_types(requested):
            if endpoints.get(alias):
                # Return the first one found in the order listed.
                return {alias: endpoints[alias]}

        # We should never get here - it's a programming logic error on our
        # part if we do. Raise this so that we can panic in unit tests.
        raise ValueError("Programming error choosing an endpoint.")

    def get_endpoints(
        self,
        service_type: str | None = None,
        interface: str | list[str] | tuple[str] | set[str] | None = None,
        region_name: str | None = None,
        service_name: str | None = None,
        service_id: str | None = None,
        endpoint_id: str | None = None,
    ) -> dict[str, list[dict[str, ty.Any] | None]]:
        """Fetch and filter endpoint data for the specified service(s).

        Returns endpoints for the specified service (or all) containing
        the specified type (or all) and region (or all) and service name.

        If there is no name in the service catalog the service_name check will
        be skipped.  This allows compatibility with services that existed
        before the name was available in the catalog.

        Returns a dict keyed by service_type with a list of endpoint dicts
        """
        endpoints_data = self.get_endpoints_data(
            service_type=service_type,
            interface=interface,
            region_name=region_name,
            service_name=service_name,
            service_id=service_id,
            endpoint_id=endpoint_id,
        )
        endpoints = {}
        for service_type, data in endpoints_data.items():
            endpoints[service_type] = self._denormalize_endpoints(data)
        return endpoints

    def get_endpoint_data_list(
        self,
        service_type: str | None = None,
        interface: str | list[str] | tuple[str] | set[str] | None = 'public',
        region_name: str | None = None,
        service_name: str | None = None,
        service_id: str | None = None,
        endpoint_id: str | None = None,
    ) -> list[discover.EndpointData]:
        """Fetch a flat list of matching EndpointData objects.

        Fetch the endpoints from the service catalog for a particular
        endpoint attribute. If no attribute is given, return the first
        endpoint of the specified type.

        Valid interface types: `public` or `publicURL`,
                               `internal` or `internalURL`,
                               `admin` or 'adminURL`

        :param string service_type: Service type of the endpoint.
        :param interface: Type of endpoint. Can be a single value or a list
                          of values. If it's a list of values, they will be
                          looked for in order of preference.
        :param string region_name: Region of the endpoint.
        :param string service_name: The assigned name of the service.
        :param string service_id: The identifier of a service.
        :param string endpoint_id: The identifier of an endpoint.

        :returns: a list of matching EndpointData objects
        :rtype: list(`keystoneauth1.discover.EndpointData`)
        """
        endpoints = self.get_endpoints_data(
            service_type=service_type,
            interface=interface,
            region_name=region_name,
            service_name=service_name,
            service_id=service_id,
            endpoint_id=endpoint_id,
        )
        return [endpoint for data in endpoints.values() for endpoint in data]

    def get_urls(
        self,
        service_type: str | None = None,
        interface: str | list[str] | tuple[str] | set[str] | None = 'public',
        region_name: str | None = None,
        service_name: str | None = None,
        service_id: str | None = None,
        endpoint_id: str | None = None,
    ) -> tuple[str | None, ...]:
        """Fetch endpoint urls from the service catalog.

        Fetch the urls of endpoints from the service catalog for a particular
        endpoint attribute. If no attribute is given, return the url of the
        first endpoint of the specified type.

        Valid interface types: `public` or `publicURL`,
                               `internal` or `internalURL`,
                               `admin` or 'adminURL`

        :param string service_type: Service type of the endpoint.
        :param interface: Type of endpoint. Can be a single value or a list
                          of values. If it's a list of values, they will be
                          looked for in order of preference.
        :param string region_name: Region of the endpoint.
        :param string service_name: The assigned name of the service.
        :param string service_id: The identifier of a service.
        :param string endpoint_id: The identifier of an endpoint.

        :returns: tuple of urls
        """
        endpoints = self.get_endpoint_data_list(
            service_type=service_type,
            interface=interface,
            region_name=region_name,
            service_name=service_name,
            service_id=service_id,
            endpoint_id=endpoint_id,
        )
        return tuple([endpoint.url for endpoint in endpoints])

    def url_for(
        self,
        service_type: str | None = None,
        interface: str | list[str] | tuple[str] | set[str] | None = 'public',
        region_name: str | None = None,
        service_name: str | None = None,
        service_id: str | None = None,
        endpoint_id: str | None = None,
    ) -> str | None:
        """Fetch an endpoint from the service catalog.

        Fetch the specified endpoint from the service catalog for
        a particular endpoint attribute. If no attribute is given, return
        the first endpoint of the specified type.

        Valid interface types: `public` or `publicURL`,
                               `internal` or `internalURL`,
                               `admin` or 'adminURL`

        :param string service_type: Service type of the endpoint.
        :param interface: Type of endpoint. Can be a single value or a list
                          of values. If it's a list of values, they will be
                          looked for in order of preference.
        :param string region_name: Region of the endpoint.
        :param string service_name: The assigned name of the service.
        :param string service_id: The identifier of a service.
        :param string endpoint_id: The identifier of an endpoint.
        """
        return self.endpoint_data_for(
            service_type=service_type,
            interface=interface,
            region_name=region_name,
            service_name=service_name,
            service_id=service_id,
            endpoint_id=endpoint_id,
        ).url

    def endpoint_data_for(
        self,
        service_type: str | None = None,
        interface: str | list[str] | tuple[str] | set[str] | None = 'public',
        region_name: str | None = None,
        service_name: str | None = None,
        service_id: str | None = None,
        endpoint_id: str | None = None,
    ) -> discover.EndpointData:
        """Fetch endpoint data from the service catalog.

        Fetch the specified endpoint data from the service catalog for
        a particular endpoint attribute. If no attribute is given, return
        the first endpoint of the specified type.

        Valid interface types: `public` or `publicURL`,
                               `internal` or `internalURL`,
                               `admin` or 'adminURL`

        :param string service_type: Service type of the endpoint.
        :param interface: Type of endpoint. Can be a single value or a list
                          of values. If it's a list of values, they will be
                          looked for in order of preference.
        :param string region_name: Region of the endpoint.
        :param string service_name: The assigned name of the service.
        :param string service_id: The identifier of a service.
        :param string endpoint_id: The identifier of an endpoint.
        """
        if not self._catalog:
            raise exceptions.EmptyCatalog('The service catalog is empty.')

        endpoint_data_list = self.get_endpoint_data_list(
            service_type=service_type,
            interface=interface,
            region_name=region_name,
            service_name=service_name,
            service_id=service_id,
            endpoint_id=endpoint_id,
        )

        if endpoint_data_list:
            return endpoint_data_list[0]

        if service_name and region_name:
            msg = (
                f'{interface} endpoint for {service_type} service '
                f'named {service_name} in {region_name} region not '
                'found'
            )
        elif service_name:
            msg = (
                f'{interface} endpoint for {service_type} service '
                f'named {service_name} not found'
            )
        elif region_name:
            msg = (
                f'{interface} endpoint for {service_type} service '
                f'in {region_name} region not found'
            )
        else:
            msg = f'{interface} endpoint for {service_type} service not found'

        raise exceptions.EndpointNotFound(msg)


class ServiceCatalogV2(ServiceCatalog):
    """An object for encapsulating the v2 service catalog.

    The object is created using raw v2 auth token from Keystone.
    """

    @classmethod
    def from_token(cls, token: dict[str, ty.Any]) -> ty_ext.Self:
        if 'access' not in token:
            raise ValueError('Invalid token format for fetching catalog')

        return cls(token['access'].get('serviceCatalog', {}))

    @staticmethod
    def normalize_interface(interface: str) -> str:
        if interface and 'URL' not in interface:
            interface += 'URL'

        return interface

    def is_interface_match(
        self, endpoint: dict[str, str], interface: str
    ) -> bool:
        return interface in endpoint

    def _normalize_endpoints(
        self, endpoints: list[dict[str, ty.Any]]
    ) -> list[dict[str, ty.Any]]:
        """Translate endpoint description dicts into v3 form.

        Takes a raw endpoint description from the catalog and changes
        it to be in v3 format. It also saves a copy of the data in
        raw_endpoint so that it can be returned by methods that expect the
        actual original data.

        :param list endpoints: List of endpoint description dicts

        :returns: List of endpoint description dicts in v3 format
        """
        new_endpoints = []
        for endpoint in endpoints:
            raw_endpoint = endpoint.copy()
            interface_urls = {}
            interface_keys = [
                key for key in endpoint.keys() if key.endswith('URL')
            ]
            for key in interface_keys:
                interface = self.normalize_interface(key)
                interface_urls[interface] = endpoint.pop(key)
            for interface, url in interface_urls.items():
                new_endpoint = endpoint.copy()
                new_endpoint['interface'] = interface
                new_endpoint['url'] = url
                # Save the actual endpoint for ease of later reconstruction
                new_endpoint['raw_endpoint'] = raw_endpoint
                new_endpoints.append(new_endpoint)
        return new_endpoints

    def _denormalize_endpoints(
        self, endpoints: list[discover.EndpointData]
    ) -> list[dict[str, ty.Any] | None]:
        """Return original endpoint description dicts.

        Takes a list of EndpointData objects and returns the original
        dict that was returned from the catalog.

        :param list endpoints: List of `keystoneauth1.discover.EndpointData`

        :returns: List of endpoint description dicts in original catalog format
        """
        raw_endpoints = super()._denormalize_endpoints(endpoints)
        # The same raw endpoint content will be in the list once for each
        # v2 endpoint_type entry. We only need one of them in the resulting
        # list. So keep a list of the string versions.
        seen = {}
        ret = []
        for endpoint in raw_endpoints:
            if str(endpoint) in seen:
                continue
            seen[str(endpoint)] = True
            ret.append(endpoint)
        return ret


class ServiceCatalogV3(ServiceCatalog):
    """An object for encapsulating the v3 service catalog.

    The object is created using raw v3 auth token from Keystone.
    """

    @classmethod
    def from_token(cls, token: dict[str, ty.Any]) -> ty_ext.Self:
        if 'token' not in token:
            raise ValueError('Invalid token format for fetching catalog')

        return cls(token['token'].get('catalog', {}))

    @staticmethod
    def normalize_interface(interface: str) -> str:
        if interface:
            interface = interface.rstrip('URL')

        return interface

    def is_interface_match(
        self, endpoint: dict[str, str], interface: str
    ) -> bool:
        try:
            return interface == endpoint['interface']
        except KeyError:
            return False



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import typing as ty

try:
    # explicitly re-export symbol
    # https://mypy.readthedocs.io/en/stable/command_line.html#cmdoption-mypy-no-implicit-reexport
    from lxml import etree as etree
except ImportError:
    etree = None  # type: ignore[assignment]

import requests
import requests.auth

from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1.identity import v3
from keystoneauth1 import session as ks_session

_PAOS_NAMESPACE = 'urn:liberty:paos:2003-08'
_ECP_NAMESPACE = 'urn:oasis:names:tc:SAML:2.0:profiles:SSO:ecp'
_PAOS_HEADER = 'application/vnd.paos+xml'

_PAOS_VER = f'ver="{_PAOS_NAMESPACE}";"{_ECP_NAMESPACE}"'

_XML_NAMESPACES = {
    'ecp': _ECP_NAMESPACE,
    'S': 'http://schemas.xmlsoap.org/soap/envelope/',
    'paos': _PAOS_NAMESPACE,
}

_XBASE = '/S:Envelope/S:Header/'
_XPATH_SP_RELAY_STATE = '//ecp:RelayState'
_XPATH_SP_CONSUMER_URL = _XBASE + 'paos:Request/@responseConsumerURL'
_XPATH_IDP_CONSUMER_URL = _XBASE + 'ecp:Response/@AssertionConsumerServiceURL'

_SOAP_FAULT = """
    <S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
       <S:Body>
         <S:Fault>
            <faultcode>S:Server</faultcode>
            <faultstring>responseConsumerURL from SP and
            assertionConsumerServiceURL from IdP do not match
            </faultstring>
         </S:Fault>
       </S:Body>
    </S:Envelope>
"""


class SamlException(Exception):
    """Base SAML plugin exception."""


class InvalidResponse(SamlException):
    """Invalid Response from SAML authentication."""


class ConsumerMismatch(SamlException):
    """The SP and IDP consumers do not match."""


def _response_xml(response: requests.Response, name: str) -> etree._Element:
    try:
        return etree.XML(response.content)
    except etree.XMLSyntaxError as e:
        msg = f'SAML2: Error parsing XML returned from {name}: {e}'
        raise InvalidResponse(msg)


def _str_from_xml(xml: etree._Element, path: str) -> str:
    li = xml.xpath(path, namespaces=_XML_NAMESPACES)
    if len(li) != 1:
        raise IndexError(f'{path} should provide a single element list')
    result = li[0]
    return str(result)  # case from _ElementUnicodeResult


_PreparedRequestT = ty.TypeVar(
    '_PreparedRequestT', bound=requests.PreparedRequest
)


class _SamlAuth(requests.auth.AuthBase):
    """A generic SAML ECP plugin for requests.

    This is a multi-step process including multiple HTTP requests.
    Authentication consists of:

    * HTTP GET request to the Service Provider.

        It's crucial to include HTTP headers indicating we are expecting SOAP
        message in return. Service Provider should respond with a SOAP
        message.

    * HTTP POST request to the external Identity Provider service with
        ECP extension enabled. The content sent is a header removed SOAP
        message returned from the Service Provider. It's also worth noting
        that ECP extension to the SAML2 doesn't define authentication method.
        The most popular is HttpBasicAuth with just user and password.
        Other possibilities could be X509 certificates or Kerberos.
        Upon successful authentication the user should receive a SAML2
        assertion.

    * HTTP POST request again to the Service Provider. The body of the
        request includes SAML2 assertion issued by a trusted Identity
        Provider. The request should be sent to the Service Provider
        consumer url specified in the SAML2 assertion.
        Providing the authentication was successful and both Service Provider
        and Identity Providers are trusted to each other, the Service
        Provider will issue an unscoped token with a list of groups the
        federated user is a member of.
    """

    def __init__(
        self,
        identity_provider_url: str,
        requests_auth: None | requests.auth.AuthBase | tuple[str, str],
    ):
        super().__init__()
        self.identity_provider_url = identity_provider_url
        self.requests_auth = requests_auth

    def __call__(self, request: _PreparedRequestT) -> _PreparedRequestT:
        try:
            accept = request.headers['Accept']
        except KeyError:
            request.headers['Accept'] = _PAOS_HEADER
        else:
            request.headers['Accept'] = ','.join([accept, _PAOS_HEADER])

        request.headers['PAOS'] = _PAOS_VER
        request.register_hook('response', self._handle_response)  # type: ignore[no-untyped-call]
        return request

    def _handle_response(
        self, response: requests.Response, **kwargs: ty.Any
    ) -> requests.Response:
        if (
            response.status_code == 200
            and response.headers.get('Content-Type') == _PAOS_HEADER
        ):
            response = self._ecp_retry(response, **kwargs)

        return response

    def _ecp_retry(
        self, sp_response: requests.Response, **kwargs: ty.Any
    ) -> requests.Response:
        history = [sp_response]

        def send(
            method: str,
            url: str,
            headers: ty.Mapping[str, str],
            data: bytes,
            auth: None | requests.auth.AuthBase | tuple[str, str] = None,
            cookies: requests.cookies.RequestsCookieJar | None = None,
        ) -> requests.Response:
            req = requests.Request(
                method=method,
                url=url,
                headers=headers,
                data=data,
                auth=auth,
                cookies=cookies,
            )
            return sp_response.connection.send(req.prepare(), **kwargs)

        authn_request = _response_xml(sp_response, 'Service Provider')
        relay_state = authn_request.xpath(
            _XPATH_SP_RELAY_STATE, namespaces=_XML_NAMESPACES
        )[0]
        sp_consumer_url = _str_from_xml(authn_request, _XPATH_SP_CONSUMER_URL)

        authn_request.remove(authn_request[0])

        idp_response = send(
            'POST',
            self.identity_provider_url,
            headers={'Content-type': 'text/xml'},
            data=etree.tostring(authn_request),
            auth=self.requests_auth,
        )
        history.append(idp_response)

        authn_response = _response_xml(idp_response, 'Identity Provider')
        idp_consumer_url = _str_from_xml(
            authn_response, _XPATH_IDP_CONSUMER_URL
        )

        if sp_consumer_url != idp_consumer_url:
            # send fault message to the SP, discard the response
            send(
                'POST',
                sp_consumer_url,
                data=_SOAP_FAULT.encode('utf-8'),
                headers={'Content-Type': _PAOS_HEADER},
            )

            # prepare error message and raise an exception.
            msg = (
                'Consumer URLs from Service Provider %(service_provider)s '
                '%(sp_consumer_url)s and Identity Provider '
                '%(identity_provider)s %(idp_consumer_url)s are not equal'
            )
            msg = msg % {
                'service_provider': sp_response.request.url,
                'sp_consumer_url': sp_consumer_url,
                'identity_provider': self.identity_provider_url,
                'idp_consumer_url': idp_consumer_url,
            }

            raise ConsumerMismatch(msg)

        # FIXME(stephenfin): We need a better type here
        authn_response[0][0] = relay_state

        # idp_consumer_url is the URL on the SP that handles the ECP body
        # returned and creates an authenticated session.
        final_resp = send(
            'POST',
            idp_consumer_url,
            headers={'Content-Type': _PAOS_HEADER},
            cookies=idp_response.cookies,
            data=etree.tostring(authn_response),
        )

        history.append(final_resp)

        # the SP should then redirect us back to the original URL to retry the
        # original request.
        if final_resp.status_code in (
            requests.codes.found,
            requests.codes.other,
        ):
            # Consume content and release the original connection
            # to allow our new request to reuse the same one.
            sp_response.content
            sp_response.raw.release_conn()

            req = sp_response.request.copy()
            req.url = final_resp.headers['location']
            req.prepare_cookies(final_resp.cookies)

            final_resp = sp_response.connection.send(req, **kwargs)
            history.append(final_resp)

        final_resp.history.extend(history)
        return final_resp


class _FederatedSaml(v3.FederationBaseAuth):
    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        protocol: str,
        identity_provider_url: str,
        *,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        super().__init__(
            auth_url,
            identity_provider,
            protocol,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
        self.identity_provider_url = identity_provider_url

    @abc.abstractmethod
    def get_requests_auth(self) -> requests.auth.AuthBase:
        raise NotImplementedError()

    def get_unscoped_auth_ref(
        self, session: ks_session.Session
    ) -> access.AccessInfoV3:
        method = self.get_requests_auth()
        auth = _SamlAuth(self.identity_provider_url, method)

        try:
            resp = session.get(
                self.federated_token_url,
                requests_auth=auth,
                authenticated=False,
            )
        except SamlException as e:
            raise exceptions.AuthorizationFailure(str(e))

        access_info = access.create(resp=resp)
        assert isinstance(access_info, access.AccessInfoV3)  # nosec B101
        return access_info


class Password(_FederatedSaml):
    r"""Implement authentication plugin for SAML2 protocol.

    ECP stands for `Enhanced Client or Proxy` and is a SAML2 extension
    for federated authentication where a transportation layer consists of
    HTTP protocol and XML SOAP messages.

    `Read for more information
    <https://wiki.shibboleth.net/confluence/display/CONCEPT/ECP>`_ on ECP.

    Reference the `SAML2 ECP specification <https://www.oasis-open.org/\
    committees/download.php/49979/saml-ecp-v2.0-wd09.pdf>`_.

    Currently only HTTPBasicAuth mechanism is available for the IdP
    authenication.

    :param auth_url: URL of the Identity Service
    :type auth_url: string

    :param identity_provider: name of the Identity Provider the client will
                              authenticate against. This parameter will be used
                              to build a dynamic URL used to obtain unscoped
                              OpenStack token.
    :type identity_provider: string

    :param identity_provider_url: An Identity Provider URL, where the SAML2
                                  authn request will be sent.
    :type identity_provider_url: string

    :param username: User's login
    :type username: string

    :param password: User's password
    :type password: string

    :param protocol: Protocol to be used for the authentication.
                     The name must be equal to one configured at the
                     keystone sp side. This value is used for building
                     dynamic authentication URL.
                     Typical value  would be: saml2
    :type protocol: string

    """

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        protocol: str,
        identity_provider_url: str,
        username: str,
        password: str,
        *,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        super().__init__(
            auth_url,
            identity_provider,
            protocol,
            identity_provider_url,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
        self.username = username
        self.password = password

    def get_requests_auth(self) -> requests.auth.AuthBase:
        return requests.auth.HTTPBasicAuth(self.username, self.password)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from keystoneauth1 import loading
from keystoneauth1.loading import opts
from keystoneauth1 import token_endpoint


class AdminToken(loading.BaseLoader[token_endpoint.Token]):
    """Authenticate with an existing token and a known endpoint.

    This plugin is primarily useful for development or for use with identity
    service ADMIN tokens. Because this token is used directly there is no
    fetching a service catalog or determining scope information and so it
    cannot be used by clients that expect use this scope information.

    Because there is no service catalog the endpoint that is supplied with
    initialization is used for all operations performed with this plugin so
    must be the full base URL to an actual service.
    """

    @property
    def plugin_class(self) -> type[token_endpoint.Token]:
        return token_endpoint.Token

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'endpoint',
                    deprecated=[loading.Opt('url')],
                    help='The endpoint that will always be used',
                ),
                loading.Opt(
                    'token',
                    secret=True,
                    help='The token that will always be used',
                ),
            ]
        )

        return options



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import logging

import iso8601


def get_logger(name: str) -> logging.Logger:
    name = name.replace(__name__.split('.')[0], 'keystoneauth')
    return logging.getLogger(name)


logger = get_logger(__name__)


def normalize_time(timestamp: datetime.datetime) -> datetime.datetime:
    """Normalize time in arbitrary timezone to UTC naive object."""
    offset = timestamp.utcoffset()
    if offset is None:
        return timestamp
    return timestamp.replace(tzinfo=None) - offset


def parse_isotime(timestr: str) -> datetime.datetime:
    """Parse time from ISO 8601 format."""
    try:
        return iso8601.parse_date(timestr)
    except iso8601.ParseError as e:
        raise ValueError(str(e))
    except TypeError as e:
        raise ValueError(str(e))


def from_utcnow(
    days: int | float = 0,
    seconds: int | float = 0,
    microseconds: int | float = 0,
    milliseconds: int | float = 0,
    minutes: int | float = 0,
    hours: int | float = 0,
    weeks: int | float = 0,
) -> datetime.datetime:
    """Calculate the time in the future from utcnow.

    :param days: Days to add to timestamp.
    :param seconds: Seconds to add to timestamp.
    :param microseconds: Microseconds to add to timestamp.
    :param milliseconds: Milliseconds to add to timestamp.
    :param minutes: Minutes to add to timestamp.
    :param hours: Hours to add to timestamp.
    :param weeks: Weeks to add to timestamp.
    :returns:
        The time in the future based on ``timedelta_kwargs`` and in TZ-naive
        format.
    :rtype:
        datetime.datetime
    """
    now = datetime.datetime.now(datetime.UTC).replace(tzinfo=None)
    delta = datetime.timedelta(
        days, seconds, microseconds, milliseconds, minutes, hours, weeks
    )
    return now + delta


def before_utcnow(
    days: int | float = 0,
    seconds: int | float = 0,
    microseconds: int | float = 0,
    milliseconds: int | float = 0,
    minutes: int | float = 0,
    hours: int | float = 0,
    weeks: int | float = 0,
) -> datetime.datetime:
    r"""Calculate the time in the past from utcnow.

    :param days: Days to remove from timestamp.
    :param seconds: Seconds to remove from timestamp.
    :param microseconds: Microseconds to remove from timestamp.
    :param milliseconds: Milliseconds to remove from timestamp.
    :param minutes: Minutes to remove from timestamp.
    :param hours: Hours to remove from timestamp.
    :param weeks: Weeks to remove from timestamp.
    :returns:
        The time in the past based on ``timedelta_kwargs`` and in TZ-naive
        format.
    :rtype:
        datetime.datetime
    """
    now = datetime.datetime.now(datetime.UTC).replace(tzinfo=None)
    delta = datetime.timedelta(
        days, seconds, microseconds, milliseconds, minutes, hours, weeks
    )
    return now - delta


# Detect if running on the Windows Subsystem for Linux
try:
    with open('/proc/version') as f:
        is_windows_linux_subsystem = 'microsoft' in f.read().lower()
except OSError:
    is_windows_linux_subsystem = False



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import datetime
import json
import uuid

from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1 import fixture
from keystoneauth1 import identity
from keystoneauth1.identity import v3
from keystoneauth1 import session
from keystoneauth1.tests.unit import k2k_fixtures
from keystoneauth1.tests.unit import utils


class TesterFederationPlugin(v3.FederationBaseAuth):
    def get_unscoped_auth_ref(self, sess):
        # This would go and talk to an idp or something
        resp = sess.post(self.federated_token_url, authenticated=False)
        return access.create(resp=resp)


class V3FederatedPluginBase(utils.TestCase):
    """Fixtures for a federated plugin, without any tests of its own.

    Separate from the tests so that another test class can reuse the setup
    without also inheriting and rerunning them.
    """

    AUTH_URL = 'http://keystone/v3'

    def setUp(self):
        super().setUp()

        self.unscoped_token = fixture.V3Token()
        self.unscoped_token_id = uuid.uuid4().hex
        self.scoped_token = copy.deepcopy(self.unscoped_token)
        self.scoped_token.set_project_scope()
        self.scoped_token.methods.append('token')
        self.scoped_token_id = uuid.uuid4().hex

        s = self.scoped_token.add_service('compute', name='nova')
        s.add_standard_endpoints(
            public='http://nova/public',
            admin='http://nova/admin',
            internal='http://nova/internal',
        )

        self.idp = uuid.uuid4().hex
        self.protocol = uuid.uuid4().hex

        self.token_url = (
            f'{self.AUTH_URL}/OS-FEDERATION/identity_providers/{self.idp}/protocols/{self.protocol}'
            '/auth'
        )

        headers = {'X-Subject-Token': self.unscoped_token_id}
        self.unscoped_mock = self.requests_mock.post(
            self.token_url, json=self.unscoped_token, headers=headers
        )

        headers = {'X-Subject-Token': self.scoped_token_id}
        auth_url = self.AUTH_URL + '/auth/tokens'
        self.scoped_mock = self.requests_mock.post(
            auth_url, json=self.scoped_token, headers=headers
        )

    def get_plugin(self, **kwargs):
        kwargs.setdefault('auth_url', self.AUTH_URL)
        kwargs.setdefault('protocol', self.protocol)
        kwargs.setdefault('identity_provider', self.idp)
        return TesterFederationPlugin(**kwargs)


class V3FederatedPlugin(V3FederatedPluginBase):
    def test_federated_url(self):
        plugin = self.get_plugin()
        self.assertEqual(self.token_url, plugin.federated_token_url)

    def test_federated_url_unversioned(self):
        plugin = self.get_plugin(auth_url="http://keystone/")
        self.assertEqual(self.token_url, plugin.federated_token_url)

    def test_unscoped_behaviour(self):
        sess = session.Session(auth=self.get_plugin())
        self.assertEqual(self.unscoped_token_id, sess.get_token())

        self.assertTrue(self.unscoped_mock.called)
        self.assertFalse(self.scoped_mock.called)

    def test_scoped_behaviour(self):
        auth = self.get_plugin(project_id=self.scoped_token.project_id)
        sess = session.Session(auth=auth)
        self.assertEqual(self.scoped_token_id, sess.get_token())

        self.assertTrue(self.unscoped_mock.called)
        self.assertTrue(self.scoped_mock.called)

    def test_scoped_behaviour_for_each_scope(self):
        # Every scope the plugin accepts has to reach the rescoping request.
        # Silently returning the unscoped token instead leaves the caller
        # authenticated as nobody in particular, which only shows up later as
        # an unexplained 403 or 404 from whatever they were trying to reach.
        scopes = [
            (
                'project',
                {
                    'project_id': uuid.uuid4().hex,
                    'project_domain_id': uuid.uuid4().hex,
                },
            ),
            ('domain', {'domain_id': uuid.uuid4().hex}),
            ('system', {'system_scope': 'all'}),
            ('trust', {'trust_id': uuid.uuid4().hex}),
        ]

        for description, kwargs in scopes:
            with self.subTest(description):
                sess = session.Session(auth=self.get_plugin(**kwargs))

                self.assertEqual(self.scoped_token_id, sess.get_token())

                scope = self.scoped_mock.last_request.json()['auth']['scope']
                self.assertIsNotNone(scope)


class K2KAuthPluginTest(utils.TestCase):
    TEST_ROOT_URL = 'http://127.0.0.1:5000/'
    TEST_URL = '{}{}'.format(TEST_ROOT_URL, 'v3')
    TEST_PASS = 'password'

    REQUEST_ECP_URL = TEST_URL + '/auth/OS-FEDERATION/saml2/ecp'

    SP_ROOT_URL = 'https://sp1.com/v3'
    SP_ID = 'sp1'
    SP_URL = 'https://sp1.com/Shibboleth.sso/SAML2/ECP'
    SP_AUTH_URL = (
        SP_ROOT_URL + '/OS-FEDERATION/identity_providers'
        '/testidp/protocols/saml2/auth'
    )

    SERVICE_PROVIDER_DICT = {
        'id': SP_ID,
        'auth_url': SP_AUTH_URL,
        'sp_url': SP_URL,
    }

    def setUp(self):
        super().setUp()
        self.token_v3 = fixture.V3Token()
        self.token_v3.add_service_provider(
            self.SP_ID, self.SP_AUTH_URL, self.SP_URL
        )
        self.session = session.Session()

        self.k2kplugin = self.get_plugin()

    def _get_base_plugin(self):
        self.stub_url(
            'POST',
            ['auth', 'tokens'],
            headers={'X-Subject-Token': uuid.uuid4().hex},
            json=self.token_v3,
        )
        return v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )

    def _mock_k2k_flow_urls(self, redirect_code=302):
        # List versions available for auth
        self.requests_mock.get(
            self.TEST_URL,
            json={'version': fixture.V3Discovery(self.TEST_URL)},
            headers={'Content-Type': 'application/json'},
        )

        # The IdP should return a ECP wrapped assertion when requested
        self.requests_mock.register_uri(
            'POST',
            self.REQUEST_ECP_URL,
            content=bytes(k2k_fixtures.ECP_ENVELOPE, 'latin-1'),
            headers={'Content-Type': 'application/vnd.paos+xml'},
            status_code=200,
        )

        # The SP should respond with a redirect (302 or 303)
        self.requests_mock.register_uri(
            'POST',
            self.SP_URL,
            content=bytes(k2k_fixtures.TOKEN_BASED_ECP, 'latin-1'),
            headers={'Content-Type': 'application/vnd.paos+xml'},
            status_code=redirect_code,
        )

        # Should not follow the redirect URL, but use the auth_url attribute
        self.requests_mock.register_uri(
            'GET',
            self.SP_AUTH_URL,
            json=k2k_fixtures.UNSCOPED_TOKEN,
            headers={'X-Subject-Token': k2k_fixtures.UNSCOPED_TOKEN_HEADER},
        )

    def get_plugin(self, **kwargs):
        kwargs.setdefault('base_plugin', self._get_base_plugin())
        kwargs.setdefault('service_provider', self.SP_ID)
        return v3.Keystone2Keystone(**kwargs)

    def test_remote_url(self):
        remote_auth_url = self.k2kplugin._remote_auth_url(self.SP_AUTH_URL)
        self.assertEqual(self.SP_ROOT_URL, remote_auth_url)

    def test_fail_getting_ecp_assertion(self):
        self.requests_mock.get(
            self.TEST_URL,
            json={'version': fixture.V3Discovery(self.TEST_URL)},
            headers={'Content-Type': 'application/json'},
        )

        self.requests_mock.register_uri(
            'POST', self.REQUEST_ECP_URL, status_code=401
        )

        self.assertRaises(
            exceptions.AuthorizationFailure,
            self.k2kplugin._get_ecp_assertion,
            self.session,
        )

    def test_get_ecp_assertion_empty_response(self):
        self.requests_mock.get(
            self.TEST_URL,
            json={'version': fixture.V3Discovery(self.TEST_URL)},
            headers={'Content-Type': 'application/json'},
        )

        self.requests_mock.register_uri(
            'POST',
            self.REQUEST_ECP_URL,
            headers={'Content-Type': 'application/vnd.paos+xml'},
            content=b'',
            status_code=200,
        )

        self.assertRaises(
            exceptions.InvalidResponse,
            self.k2kplugin._get_ecp_assertion,
            self.session,
        )

    def test_get_ecp_assertion_wrong_headers(self):
        self.requests_mock.get(
            self.TEST_URL,
            json={'version': fixture.V3Discovery(self.TEST_URL)},
            headers={'Content-Type': 'application/json'},
        )

        self.requests_mock.register_uri(
            'POST',
            self.REQUEST_ECP_URL,
            headers={'Content-Type': uuid.uuid4().hex},
            content=b'',
            status_code=200,
        )

        self.assertRaises(
            exceptions.InvalidResponse,
            self.k2kplugin._get_ecp_assertion,
            self.session,
        )

    def test_send_ecp_authn_response(self):
        self._mock_k2k_flow_urls()
        # Perform the request
        response = self.k2kplugin._send_service_provider_ecp_authn_response(
            self.session, self.SP_URL, self.SP_AUTH_URL
        )

        # Check the response
        self.assertEqual(
            k2k_fixtures.UNSCOPED_TOKEN_HEADER,
            response.headers['X-Subject-Token'],
        )

    def test_send_ecp_authn_response_303_redirect(self):
        self._mock_k2k_flow_urls(redirect_code=303)
        # Perform the request
        response = self.k2kplugin._send_service_provider_ecp_authn_response(
            self.session, self.SP_URL, self.SP_AUTH_URL
        )

        # Check the response
        self.assertEqual(
            k2k_fixtures.UNSCOPED_TOKEN_HEADER,
            response.headers['X-Subject-Token'],
        )

    def test_end_to_end_workflow(self):
        self._mock_k2k_flow_urls()
        auth_ref = self.k2kplugin.get_auth_ref(self.session)
        self.assertEqual(
            k2k_fixtures.UNSCOPED_TOKEN_HEADER, auth_ref.auth_token
        )

    def test_end_to_end_workflow_303_redirect(self):
        self._mock_k2k_flow_urls(redirect_code=303)
        auth_ref = self.k2kplugin.get_auth_ref(self.session)
        self.assertEqual(
            k2k_fixtures.UNSCOPED_TOKEN_HEADER, auth_ref.auth_token
        )

    def test_end_to_end_with_generic_password(self):
        # List versions available for auth
        self.requests_mock.get(
            self.TEST_ROOT_URL,
            json=fixture.DiscoveryList(self.TEST_ROOT_URL),
            headers={'Content-Type': 'application/json'},
        )

        # The IdP should return a ECP wrapped assertion when requested
        self.requests_mock.register_uri(
            'POST',
            self.REQUEST_ECP_URL,
            content=bytes(k2k_fixtures.ECP_ENVELOPE, 'latin-1'),
            headers={'Content-Type': 'application/vnd.paos+xml'},
            status_code=200,
        )

        # The SP should respond with a redirect (302 or 303)
        self.requests_mock.register_uri(
            'POST',
            self.SP_URL,
            content=bytes(k2k_fixtures.TOKEN_BASED_ECP, 'latin-1'),
            headers={'Content-Type': 'application/vnd.paos+xml'},
            status_code=302,
        )

        # Should not follow the redirect URL, but use the auth_url attribute
        self.requests_mock.register_uri(
            'GET',
            self.SP_AUTH_URL,
            json=k2k_fixtures.UNSCOPED_TOKEN,
            headers={'X-Subject-Token': k2k_fixtures.UNSCOPED_TOKEN_HEADER},
        )

        self.stub_url(
            'POST',
            ['auth', 'tokens'],
            headers={'X-Subject-Token': uuid.uuid4().hex},
            json=self.token_v3,
        )

        plugin = identity.Password(
            self.TEST_ROOT_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            user_domain_id='default',
        )

        k2kplugin = self.get_plugin(base_plugin=plugin)
        self.assertEqual(
            k2k_fixtures.UNSCOPED_TOKEN_HEADER,
            k2kplugin.get_token(self.session),
        )


class V3FederatedUnscopedState(V3FederatedPluginBase):
    """The API a caller uses to hold on to the unscoped token itself."""

    def test_state_is_empty_before_authenticating(self):
        self.assertIsNone(self.get_plugin().get_unscoped_auth_state())

    def test_one_unscoped_token_serves_every_scope(self):
        origin = self.get_plugin()
        session.Session(auth=origin).get_token()
        state = origin.get_unscoped_auth_state()

        self.assertIsNotNone(state)
        self.assertEqual(1, self.unscoped_mock.call_count)

        scopes = [
            {'project_id': self.scoped_token.project_id},
            {'domain_id': uuid.uuid4().hex},
            {'system_scope': 'all'},
            {'trust_id': uuid.uuid4().hex},
        ]

        for kwargs in scopes:
            with self.subTest(next(iter(kwargs))):
                plugin = self.get_plugin(**kwargs)
                plugin.set_unscoped_auth_state(state)

                self.assertEqual(
                    self.scoped_token_id,
                    session.Session(auth=plugin).get_token(),
                )

        # One authentication, and a rescope for each of the scopes.
        self.assertEqual(1, self.unscoped_mock.call_count)
        self.assertEqual(len(scopes), self.scoped_mock.call_count)

    def test_cache_id_ignores_the_scope_but_not_the_identity(self):
        base_id = self.get_plugin().get_unscoped_cache_id()
        self.assertIsNotNone(base_id)

        for kwargs in (
            {'project_id': uuid.uuid4().hex},
            {'domain_id': uuid.uuid4().hex},
            {'system_scope': 'all'},
            {'trust_id': uuid.uuid4().hex},
        ):
            with self.subTest(f'unchanged by {next(iter(kwargs))}'):
                self.assertEqual(
                    base_id, self.get_plugin(**kwargs).get_unscoped_cache_id()
                )

        for description, kwargs in (
            ('identity provider', {'identity_provider': uuid.uuid4().hex}),
            ('protocol', {'protocol': uuid.uuid4().hex}),
            ('auth url', {'auth_url': 'http://elsewhere/v3'}),
            # 'foo'/'bar' against 'foob'/'ar' needs the elements terminated
            # rather than simply concatenated.
            (
                'element boundaries',
                {'identity_provider': 'foo', 'protocol': 'bar'},
            ),
        ):
            with self.subTest(f'changed by {description}'):
                self.assertNotEqual(
                    base_id, self.get_plugin(**kwargs).get_unscoped_cache_id()
                )

    def test_expired_state_is_replaced(self):
        expired = fixture.V3Token()
        expired.expires = datetime.datetime.now(
            datetime.UTC
        ) - datetime.timedelta(hours=1)
        plugin = self.get_plugin(project_id=self.scoped_token.project_id)
        plugin.set_unscoped_auth_state(
            json.dumps({'auth_token': uuid.uuid4().hex, 'body': expired})
        )

        session.Session(auth=plugin).get_token()

        self.assertEqual(1, self.unscoped_mock.call_count)

    def test_a_refused_unscoped_token_is_replaced(self):
        # A revoked token is not expired, so nothing about it says it will be
        # refused until the rescope is. The rescope carries the token in its
        # body rather than a header, so the session cannot retry it either.
        plugin = self.get_plugin(project_id=self.scoped_token.project_id)
        plugin.set_unscoped_auth_state(
            json.dumps({'auth_token': 'revoked', 'body': fixture.V3Token()})
        )

        def rescope(request, context):
            token = request.json()['auth']['identity']['token']['id']
            if token == 'revoked':
                context.status_code = 401
                return {'error': {'message': 'revoked'}}
            context.status_code = 201
            context.headers['X-Subject-Token'] = self.scoped_token_id
            return self.scoped_token

        self.scoped_mock = self.requests_mock.post(
            self.AUTH_URL + '/auth/tokens', json=rescope
        )

        self.assertEqual(
            self.scoped_token_id, session.Session(auth=plugin).get_token()
        )
        self.assertEqual(1, self.unscoped_mock.call_count)

    def test_a_refused_fresh_unscoped_token_is_not_retried(self):
        # Authenticating again would only produce the same refusal, so the
        # error belongs to the caller rather than another round trip.
        self.requests_mock.post(
            self.AUTH_URL + '/auth/tokens',
            status_code=401,
            json={'error': {'message': 'nope'}},
        )
        plugin = self.get_plugin(project_id=self.scoped_token.project_id)

        self.assertRaises(
            exceptions.Unauthorized, session.Session(auth=plugin).get_token
        )
        self.assertEqual(1, self.unscoped_mock.call_count)

    def test_invalidate_discards_the_unscoped_token(self):
        plugin = self.get_plugin(project_id=self.scoped_token.project_id)
        session.Session(auth=plugin).get_token()
        self.assertIsNotNone(plugin.get_unscoped_auth_state())

        self.assertTrue(plugin.invalidate())

        # Otherwise the retry after an unauthorized response would rescope
        # from a token the identity service has already refused.
        self.assertIsNone(plugin.get_unscoped_auth_state())

    def test_state_that_did_not_come_from_here_is_refused(self):
        plugin = self.get_plugin()

        for data in (
            '{}',
            '{"auth_token": "x"}',
            '{"body": {"access": {}}}',
            # Valid JSON, but not the object shape this API produces. These
            # come back from persistent storage, where stale or malformed
            # data is possible, so they must be a ValueError rather than the
            # TypeError the key lookup would otherwise raise.
            '[]',
            'null',
            '"x"',
        ):
            with self.subTest(data):
                self.assertRaises(
                    ValueError, plugin.set_unscoped_auth_state, data
                )

    def test_empty_state_clears(self):
        plugin = self.get_plugin(project_id=self.scoped_token.project_id)
        session.Session(auth=plugin).get_token()

        plugin.set_unscoped_auth_state(None)

        self.assertIsNone(plugin.get_unscoped_auth_state())

    def test_interactive_flows_are_marked(self):
        for plugin_class, interactive in (
            (identity.V3OidcAuthorizationCode, True),
            (identity.V3OidcDeviceAuthorization, True),
            (identity.V3OidcPassword, False),
            (identity.V3OidcClientCredentials, False),
        ):
            with self.subTest(plugin_class.__name__):
                self.assertEqual(
                    interactive, plugin_class.interactive_unscoped_auth
                )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1 import fixture
from keystoneauth1 import identity
from keystoneauth1.loading._plugins.identity import generic
from keystoneauth1 import session
from keystoneauth1.tests.unit.loading import utils


class PasswordTests(utils.TestCase):
    def test_options(self):
        opts = [o.name for o in generic.Password().get_options()]

        allowed_opts = [
            'username',
            'user-domain-id',
            'user-domain-name',
            'user-id',
            'password',
            'system-scope',
            'domain-id',
            'domain-name',
            'project-id',
            'project-name',
            'project-domain-id',
            'project-domain-name',
            'trust-id',
            'auth-url',
            'default-domain-id',
            'default-domain-name',
        ]

        self.assertEqual(set(allowed_opts), set(opts))
        self.assertEqual(len(allowed_opts), len(opts))

    def test_loads_v3_with_user_domain(self):
        auth_url = 'http://keystone.test:5000'
        disc = fixture.DiscoveryList(href=auth_url)
        sess = session.Session()
        self.requests_mock.get(auth_url, json=disc)

        plugin = generic.Password().load_from_options(
            auth_url=auth_url,
            user_id=uuid.uuid4().hex,
            password=uuid.uuid4().hex,
            project_id=uuid.uuid4().hex,
            user_domain_id=uuid.uuid4().hex,
        )

        inner_plugin = plugin._do_create_plugin(sess)

        self.assertIsInstance(inner_plugin, identity.V3Password)
        self.assertEqual(inner_plugin.auth_url, auth_url + '/v3')


class TokenTests(utils.TestCase):
    def test_options(self):
        opts = [o.name for o in generic.Token().get_options()]

        allowed_opts = [
            'token',
            'system-scope',
            'domain-id',
            'domain-name',
            'project-id',
            'project-name',
            'project-domain-id',
            'project-domain-name',
            'trust-id',
            'auth-url',
            'default-domain-id',
            'default-domain-name',
        ]

        self.assertEqual(set(allowed_opts), set(opts))
        self.assertEqual(len(allowed_opts), len(opts))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from keystoneauth1 import access
from keystoneauth1.identity import base
from keystoneauth1 import session as ks_session


class AccessInfoPlugin(base.BaseIdentityPlugin):
    """A plugin that turns an existing AccessInfo object into a usable plugin.

    There are cases where reuse of an auth_ref or AccessInfo object is
    warranted such as from a cache, from auth_token middleware, or another
    source.

    Turn the existing access info object into an identity plugin. This plugin
    cannot be refreshed as the AccessInfo object does not contain any
    authorizing information.

    :param auth_ref: the existing AccessInfo object.
    :type auth_ref: keystoneauth1.access.AccessInfo
    :param auth_url: the url where this AccessInfo was retrieved from. Required
                     if using the AUTH_INTERFACE with get_endpoint. (optional)
    """

    auth_ref: access.AccessInfo

    def __init__(
        self, auth_ref: access.AccessInfo, auth_url: str | None = None
    ):
        super().__init__(auth_url=auth_url, reauthenticate=False)
        self.auth_ref = auth_ref

    def get_auth_ref(self, session: ks_session.Session) -> access.AccessInfo:
        return self.auth_ref

    def invalidate(self) -> bool:
        # NOTE(jamielennox): Don't allow the default invalidation to occur
        # because on next authentication request we will only get the same
        # auth_ref object again.
        return False



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import typing as ty

import stevedore
from stevedore import extension

from keystoneauth1 import exceptions
from keystoneauth1 import plugin

if ty.TYPE_CHECKING:
    from keystoneauth1.loading import base
    from keystoneauth1.loading import opts

PLUGIN_NAMESPACE = 'keystoneauth1.plugin'


__all__ = (
    'get_available_plugin_names',
    'get_available_plugin_loaders',
    'get_plugin_loader',
    'get_plugin_options',
    'BaseLoader',
    'PLUGIN_NAMESPACE',
)

T = ty.TypeVar('T', covariant=True)


def _auth_plugin_available(
    ext: extension.Extension['base.BaseLoader[ty.Any]'],
) -> bool:
    """Read the value of available for whether to load this plugin."""
    assert ext.obj is not None
    return ext.obj.available


def get_available_plugin_names() -> frozenset[str]:
    """Get the names of all the plugins that are available on the system.

    This is particularly useful for help and error text to prompt a user for
    example what plugins they may specify.

    :returns: A list of names.
    :rtype: frozenset
    """
    mgr = stevedore.EnabledExtensionManager(
        namespace=PLUGIN_NAMESPACE,
        check_func=_auth_plugin_available,
        invoke_on_load=True,
        propagate_map_exceptions=True,
    )
    return frozenset(mgr.names())


def get_available_plugin_loaders() -> dict[
    str, 'BaseLoader[plugin.BaseAuthPluginT]'
]:
    """Retrieve all the plugin classes available on the system.

    :returns: A dict with plugin entrypoint name as the key and the plugin
              loader as the value.
    :rtype: dict
    """
    mgr: stevedore.EnabledExtensionManager[BaseLoader[plugin.BaseAuthPluginT]]
    mgr = stevedore.EnabledExtensionManager(
        namespace=PLUGIN_NAMESPACE,
        check_func=_auth_plugin_available,
        invoke_on_load=True,
        propagate_map_exceptions=True,
    )

    # NOTE(stephenfin): We know obj is not None since we passed
    # invoke_on_load=True above. The hints in stevedore need work.
    return dict(mgr.map(lambda ext: (ext.entry_point.name, ext.obj)))  # type: ignore


def get_plugin_loader(name: str) -> 'BaseLoader[plugin.BaseAuthPluginT]':
    """Retrieve a plugin class by its entrypoint name.

    :param str name: The name of the object to get.

    :returns: An auth plugin class.
    :rtype: :py:class:`keystoneauth1.loading.BaseLoader`

    :raises keystoneauth1.exceptions.auth_plugins.NoMatchingPlugin:
        if a plugin cannot be created.
    """
    try:
        mgr: stevedore.DriverManager[BaseLoader[plugin.BaseAuthPluginT]]
        mgr = stevedore.DriverManager(
            namespace=PLUGIN_NAMESPACE, invoke_on_load=True, name=name
        )
    except RuntimeError:
        raise exceptions.NoMatchingPlugin(name)

    # NOTE(stephenfin): We know this will return an instance of the BaseLoader
    # rather than the class itself since we passed invoke_on_load=True above.
    # The hints in stevedore need work.
    return ty.cast('BaseLoader[plugin.BaseAuthPluginT]', mgr.driver)


def get_plugin_options(name: str) -> list['opts.Opt']:
    """Get the options for a specific plugin.

    This will be the list of options that is registered and loaded by the
    specified plugin.

    :returns: A list of :py:class:`keystoneauth1.loading.Opt` options.

    :raises keystoneauth1.exceptions.auth_plugins.NoMatchingPlugin:
        if a plugin cannot be created.
    """
    return get_plugin_loader(name).get_options()


class _BaseLoader(ty.Generic[T], metaclass=abc.ABCMeta):
    @property
    def plugin_class(self) -> type[T]:
        raise NotImplementedError()

    def create_plugin(self, **kwargs: ty.Any) -> T:
        """Create a plugin from the options available for the loader.

        Given the options that were specified by the loader create an
        appropriate plugin. You can override this function in your loader.

        This used to be specified by providing the plugin_class property and
        this is still supported, however specifying a property didn't let you
        choose a plugin type based upon the options that were presented.

        Override this function if you wish to return different plugins based on
        the options presented, otherwise you can simply provide the
        plugin_class property.

        Added 2.9
        """
        return self.plugin_class(**kwargs)

    @abc.abstractmethod
    def get_options(self) -> list['opts.Opt']:
        """Return the list of parameters associated with the auth plugin.

        This list may be used to generate CLI or config arguments.

        :returns: A list of Param objects describing available plugin
                  parameters.
        :rtype: list
        """
        return []

    @property
    def available(self) -> bool:
        """Return if the plugin is available for loading.

        If a plugin is missing dependencies or for some other reason should not
        be available to the current system it should override this property and
        return False to exclude itself from the plugin list.

        :rtype: bool
        """
        return True

    def load_from_options(self, **kwargs: ty.Any) -> T:
        """Create a plugin from the arguments retrieved from get_options.

        A client can override this function to do argument validation or to
        handle differences between the registered options and what is required
        to create the plugin.
        """
        missing_required = [
            o
            for o in self.get_options()
            if o.required and kwargs.get(o.dest) is None
        ]

        if missing_required:
            raise exceptions.MissingRequiredOptions(missing_required)

        return self.create_plugin(**kwargs)

    def load_from_options_getter(
        self, getter: ty.Callable[['opts.Opt'], ty.Any], **kwargs: ty.Any
    ) -> T:
        """Load a plugin from getter function that returns appropriate values.

        To handle cases other than the provided CONF and CLI loading you can
        specify a custom loader function that will be queried for the option
        value.
        The getter is a function that takes a
        :py:class:`keystoneauth1.loading.Opt` and returns a value to load with.

        :param getter: A function that returns a value for the given opt.
        :type getter: callable

        :returns: An authentication Plugin.
        :rtype: :py:class:`keystoneauth1.plugin.BaseAuthPlugin`
        """
        for opt in (o for o in self.get_options() if o.dest not in kwargs):
            val = getter(opt)
            if val is not None:
                val = opt.type(val)
            kwargs[opt.dest] = val

        return self.load_from_options(**kwargs)


class BaseLoader(_BaseLoader[plugin.BaseAuthPluginT]): ...



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import json
import time
import unittest
import uuid

from keystoneauth1 import _utils as ksa_utils
from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1.exceptions import ClientException
from keystoneauth1 import fixture
from keystoneauth1.identity import v3
from keystoneauth1.identity.v3 import base as v3_base
from keystoneauth1 import session
from keystoneauth1.tests.unit import utils


class V3IdentityPlugin(utils.TestCase):
    TEST_ROOT_URL = 'http://127.0.0.1:5000/'
    TEST_URL = '{}{}'.format(TEST_ROOT_URL, 'v3')
    TEST_ROOT_ADMIN_URL = 'http://127.0.0.1:35357/'
    TEST_ADMIN_URL = '{}{}'.format(TEST_ROOT_ADMIN_URL, 'v3')

    TEST_PASS = 'password'

    TEST_APP_CRED_ID = 'appcredid'
    TEST_APP_CRED_SECRET = 'secret'

    TEST_CLIENT_CRED_ID = 'clientcredid'
    TEST_CLIENT_CRED_SECRET = 'secret'

    TEST_SERVICE_CATALOG = [
        {
            "endpoints": [
                {
                    "url": "http://cdn.admin-nets.local:8774/v1.0/",
                    "region": "RegionOne",
                    "interface": "public",
                },
                {
                    "url": "http://127.0.0.1:8774/v1.0",
                    "region": "RegionOne",
                    "interface": "internal",
                },
                {
                    "url": "http://cdn.admin-nets.local:8774/v1.0",
                    "region": "RegionOne",
                    "interface": "admin",
                },
            ],
            "type": "nova_compat",
        },
        {
            "endpoints": [
                {
                    "url": "http://nova/novapi/public",
                    "region": "RegionOne",
                    "interface": "public",
                },
                {
                    "url": "http://nova/novapi/internal",
                    "region": "RegionOne",
                    "interface": "internal",
                },
                {
                    "url": "http://nova/novapi/admin",
                    "region": "RegionOne",
                    "interface": "admin",
                },
            ],
            "type": "compute",
            "name": "nova",
        },
        {
            "endpoints": [
                {
                    "url": "http://glance/glanceapi/public",
                    "region": "RegionOne",
                    "interface": "public",
                },
                {
                    "url": "http://glance/glanceapi/internal",
                    "region": "RegionOne",
                    "interface": "internal",
                },
                {
                    "url": "http://glance/glanceapi/admin",
                    "region": "RegionOne",
                    "interface": "admin",
                },
            ],
            "type": "image",
            "name": "glance",
        },
        {
            "endpoints": [
                {
                    "url": "http://127.0.0.1:5000/v3",
                    "region": "RegionOne",
                    "interface": "public",
                },
                {
                    "url": "http://127.0.0.1:5000/v3",
                    "region": "RegionOne",
                    "interface": "internal",
                },
                {
                    "url": TEST_ADMIN_URL,
                    "region": "RegionOne",
                    "interface": "admin",
                },
            ],
            "type": "identity",
        },
        {
            "endpoints": [
                {
                    "url": "http://swift/swiftapi/public",
                    "region": "RegionOne",
                    "interface": "public",
                },
                {
                    "url": "http://swift/swiftapi/internal",
                    "region": "RegionOne",
                    "interface": "internal",
                },
                {
                    "url": "http://swift/swiftapi/admin",
                    "region": "RegionOne",
                    "interface": "admin",
                },
            ],
            "type": "object-store",
        },
    ]

    TEST_SERVICE_PROVIDERS = [
        {
            "auth_url": "https://sp1.com/v3/OS-FEDERATION/"
            "identity_providers/acme/protocols/saml2/auth",
            "id": "sp1",
            "sp_url": "https://sp1.com/Shibboleth.sso/SAML2/ECP",
        },
        {
            "auth_url": "https://sp2.com/v3/OS-FEDERATION/"
            "identity_providers/acme/protocols/saml2/auth",
            "id": "sp2",
            "sp_url": "https://sp2.com/Shibboleth.sso/SAML2/ECP",
        },
    ]

    def setUp(self):
        super().setUp()

        self.TEST_DISCOVERY_RESPONSE = {
            'versions': {'values': [fixture.V3Discovery(self.TEST_URL)]}
        }

        nextyear = 1 + time.gmtime().tm_year
        self.TEST_RESPONSE_DICT = {
            "token": {
                "methods": ["token", "password"],
                "expires_at": f"{nextyear}-02-01T00:00:10.000123Z",
                "project": {
                    "domain": {
                        "id": self.TEST_DOMAIN_ID,
                        "name": self.TEST_DOMAIN_NAME,
                    },
                    "id": self.TEST_TENANT_ID,
                    "name": self.TEST_TENANT_NAME,
                },
                "user": {
                    "domain": {
                        "id": self.TEST_DOMAIN_ID,
                        "name": self.TEST_DOMAIN_NAME,
                    },
                    "id": self.TEST_USER,
                    "name": self.TEST_USER,
                },
                "issued_at": "2013-05-29T16:55:21.468960Z",
                "catalog": self.TEST_SERVICE_CATALOG,
                "service_providers": self.TEST_SERVICE_PROVIDERS,
            }
        }
        self.TEST_PROJECTS_RESPONSE = {
            "projects": [
                {
                    "domain_id": "1789d1",
                    "enabled": "True",
                    "id": "263fd9",
                    "links": {
                        "self": "https://identity:5000/v3/projects/263fd9"
                    },
                    "name": "Dev Group A",
                },
                {
                    "domain_id": "1789d1",
                    "enabled": "True",
                    "id": "e56ad3",
                    "links": {
                        "self": "https://identity:5000/v3/projects/e56ad3"
                    },
                    "name": "Dev Group B",
                },
            ],
            "links": {"self": "https://identity:5000/v3/projects"},
        }
        self.TEST_APP_CRED_TOKEN_RESPONSE = {
            "token": {
                "methods": ["application_credential"],
                "expires_at": f"{nextyear}-02-01T00:00:10.000123Z",
                "project": {
                    "domain": {
                        "id": self.TEST_DOMAIN_ID,
                        "name": self.TEST_DOMAIN_NAME,
                    },
                    "id": self.TEST_TENANT_ID,
                    "name": self.TEST_TENANT_NAME,
                },
                "user": {
                    "domain": {
                        "id": self.TEST_DOMAIN_ID,
                        "name": self.TEST_DOMAIN_NAME,
                    },
                    "id": self.TEST_USER,
                    "name": self.TEST_USER,
                },
                "issued_at": "2013-05-29T16:55:21.468960Z",
                "catalog": self.TEST_SERVICE_CATALOG,
                "service_providers": self.TEST_SERVICE_PROVIDERS,
                "application_credential_restricted": True,
            }
        }
        self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE = {
            "token": {
                "methods": ["oauth2_credential"],
                "expires_at": f"{nextyear}-02-01T00:00:10.000123Z",
                "project": {
                    "domain": {
                        "id": self.TEST_DOMAIN_ID,
                        "name": self.TEST_DOMAIN_NAME,
                    },
                    "id": self.TEST_TENANT_ID,
                    "name": self.TEST_TENANT_NAME,
                },
                "user": {
                    "domain": {
                        "id": self.TEST_DOMAIN_ID,
                        "name": self.TEST_DOMAIN_NAME,
                    },
                    "id": self.TEST_USER,
                    "name": self.TEST_USER,
                },
                "issued_at": "2013-05-29T16:55:21.468960Z",
                "catalog": self.TEST_SERVICE_CATALOG,
                "service_providers": self.TEST_SERVICE_PROVIDERS,
                "oauth2_credential": {
                    "x5t#S256": "7UN-z4yFIm9s4jakecGoKa4rc353pDCuFUo9fsDD_1s="
                },
            }
        }
        self.TEST_RECEIPT_RESPONSE = {
            "receipt": {
                "methods": ["password"],
                "expires_at": f"{nextyear}-02-01T00:00:10.000123Z",
                "user": {
                    "domain": {
                        "id": self.TEST_DOMAIN_ID,
                        "name": self.TEST_DOMAIN_NAME,
                    },
                    "id": self.TEST_USER,
                    "name": self.TEST_USER,
                },
                "issued_at": "2013-05-29T16:55:21.468960Z",
            },
            "required_auth_methods": [["password", "totp"]],
        }

    def stub_auth(self, subject_token=None, **kwargs):
        if not subject_token:
            subject_token = self.TEST_TOKEN

        self.stub_url(
            'POST',
            ['auth', 'tokens'],
            headers={'X-Subject-Token': subject_token},
            **kwargs,
        )

    def stub_receipt(self, receipt=None, receipt_data=None, **kwargs):
        if not receipt:
            receipt = self.TEST_RECEIPT

        if not receipt_data:
            receipt_data = self.TEST_RECEIPT_RESPONSE

        self.stub_url(
            'POST',
            ['auth', 'tokens'],
            headers={'Openstack-Auth-Receipt': receipt},
            status_code=401,
            json=receipt_data,
            **kwargs,
        )

    def test_authenticate_with_username_password(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        self.assertFalse(a.has_scope_parameters)
        s = session.Session(auth=a)

        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'identity': {
                    'methods': ['password'],
                    'password': {
                        'user': {
                            'name': self.TEST_USER,
                            'password': self.TEST_PASS,
                        }
                    },
                }
            }
        }

        self.assertRequestBodyIs(json=req)
        self.assertRequestHeaderEqual('Content-Type', 'application/json')
        self.assertRequestHeaderEqual('Accept', 'application/json')
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_authenticate_with_username_password_domain_scoped(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            domain_id=self.TEST_DOMAIN_ID,
        )
        self.assertTrue(a.has_scope_parameters)
        s = session.Session(a)

        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'identity': {
                    'methods': ['password'],
                    'password': {
                        'user': {
                            'name': self.TEST_USER,
                            'password': self.TEST_PASS,
                        }
                    },
                },
                'scope': {'domain': {'id': self.TEST_DOMAIN_ID}},
            }
        }
        self.assertRequestBodyIs(json=req)
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_authenticate_with_username_password_project_scoped(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            project_id=self.TEST_TENANT_ID,
        )
        self.assertTrue(a.has_scope_parameters)
        s = session.Session(a)

        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'identity': {
                    'methods': ['password'],
                    'password': {
                        'user': {
                            'name': self.TEST_USER,
                            'password': self.TEST_PASS,
                        }
                    },
                },
                'scope': {'project': {'id': self.TEST_TENANT_ID}},
            }
        }
        self.assertRequestBodyIs(json=req)
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)
        self.assertEqual(s.auth.auth_ref.project_id, self.TEST_TENANT_ID)

    def test_authenticate_with_token(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v3.Token(self.TEST_URL, self.TEST_TOKEN)
        s = session.Session(auth=a)

        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'identity': {
                    'methods': ['token'],
                    'token': {'id': self.TEST_TOKEN},
                }
            }
        }

        self.assertRequestBodyIs(json=req)

        self.assertRequestHeaderEqual('Content-Type', 'application/json')
        self.assertRequestHeaderEqual('Accept', 'application/json')
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_with_expired(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)

        d = copy.deepcopy(self.TEST_RESPONSE_DICT)
        d['token']['expires_at'] = '2000-01-01T00:00:10.000123Z'

        a = v3.Password(
            self.TEST_URL, username='username', password='password'
        )
        a.auth_ref = access.create(body=d)
        s = session.Session(auth=a)

        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        self.assertEqual(
            a.auth_ref._data['token']['expires_at'],
            self.TEST_RESPONSE_DICT['token']['expires_at'],
        )

    def test_with_domain_and_project_scoping(self):
        a = v3.Password(
            self.TEST_URL,
            username='username',
            password='password',
            project_id='project',
            domain_id='domain',
        )

        self.assertTrue(a.has_scope_parameters)
        self.assertRaises(exceptions.AuthorizationFailure, a.get_token, None)
        self.assertRaises(exceptions.AuthorizationFailure, a.get_headers, None)

    def test_with_trust_id(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            trust_id='trust',
        )
        self.assertTrue(a.has_scope_parameters)
        s = session.Session(a)

        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'identity': {
                    'methods': ['password'],
                    'password': {
                        'user': {
                            'name': self.TEST_USER,
                            'password': self.TEST_PASS,
                        }
                    },
                },
                'scope': {'OS-TRUST:trust': {'id': 'trust'}},
            }
        }
        self.assertRequestBodyIs(json=req)
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_with_multiple_mechanisms_factory(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        p = v3.PasswordMethod(username=self.TEST_USER, password=self.TEST_PASS)
        t = v3.TokenMethod(token='foo')
        a = v3.Auth(self.TEST_URL, [p, t], trust_id='trust')
        s = session.Session(a)

        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'identity': {
                    'methods': ['password', 'token'],
                    'password': {
                        'user': {
                            'name': self.TEST_USER,
                            'password': self.TEST_PASS,
                        }
                    },
                    'token': {'id': 'foo'},
                },
                'scope': {'OS-TRUST:trust': {'id': 'trust'}},
            }
        }
        self.assertRequestBodyIs(json=req)
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_with_multiple_mechanisms(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        p = v3.PasswordMethod(username=self.TEST_USER, password=self.TEST_PASS)
        t = v3.TokenMethod(token='foo')
        a = v3.Auth(self.TEST_URL, [p, t], trust_id='trust')
        self.assertTrue(a.has_scope_parameters)
        s = session.Session(auth=a)

        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'identity': {
                    'methods': ['password', 'token'],
                    'password': {
                        'user': {
                            'name': self.TEST_USER,
                            'password': self.TEST_PASS,
                        }
                    },
                    'token': {'id': 'foo'},
                },
                'scope': {'OS-TRUST:trust': {'id': 'trust'}},
            }
        }
        self.assertRequestBodyIs(json=req)
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_with_multiple_scopes(self):
        s = session.Session()

        a = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            domain_id='x',
            project_id='x',
        )
        self.assertRaises(exceptions.AuthorizationFailure, a.get_auth_ref, s)

        a = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            domain_id='x',
            trust_id='x',
        )
        self.assertRaises(exceptions.AuthorizationFailure, a.get_auth_ref, s)

    def test_application_credential_method(self):
        self.stub_auth(json=self.TEST_APP_CRED_TOKEN_RESPONSE)
        ac = v3.ApplicationCredential(
            self.TEST_URL,
            application_credential_id=self.TEST_APP_CRED_ID,
            application_credential_secret=self.TEST_APP_CRED_SECRET,
        )
        req = {
            'auth': {
                'identity': {
                    'methods': ['application_credential'],
                    'application_credential': {
                        'id': self.TEST_APP_CRED_ID,
                        'secret': self.TEST_APP_CRED_SECRET,
                    },
                }
            }
        }
        s = session.Session(auth=ac)
        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )
        self.assertRequestBodyIs(json=req)
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def _do_service_url_test(self, base_url, endpoint_filter):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        self.stub_url(
            'GET', ['path'], base_url=base_url, text='SUCCESS', status_code=200
        )

        a = v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        resp = s.get('/path', endpoint_filter=endpoint_filter)

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(
            self.requests_mock.last_request.url, base_url + '/path'
        )

    def test_service_url(self):
        endpoint_filter = {
            'service_type': 'compute',
            'interface': 'admin',
            'service_name': 'nova',
        }
        self._do_service_url_test('http://nova/novapi/admin', endpoint_filter)

    def test_service_url_defaults_to_public(self):
        endpoint_filter = {'service_type': 'compute'}
        self._do_service_url_test('http://nova/novapi/public', endpoint_filter)

    def test_endpoint_filter_without_service_type_fails(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)

        a = v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        self.assertRaises(
            exceptions.EndpointNotFound,
            s.get,
            '/path',
            endpoint_filter={'interface': 'admin'},
        )

    def test_full_url_overrides_endpoint_filter(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        self.stub_url(
            'GET',
            [],
            base_url='http://testurl/',
            text='SUCCESS',
            status_code=200,
        )

        a = v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        resp = s.get(
            'http://testurl/', endpoint_filter={'service_type': 'compute'}
        )
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.text, 'SUCCESS')

    def test_service_providers_urls(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session()
        auth_ref = a.get_auth_ref(s)

        service_providers = auth_ref.service_providers
        self.assertEqual(
            'https://sp1.com/v3/OS-FEDERATION/'
            'identity_providers/acme/protocols/saml2/auth',
            service_providers.get_auth_url('sp1'),
        )
        self.assertEqual(
            'https://sp1.com/Shibboleth.sso/SAML2/ECP',
            service_providers.get_sp_url('sp1'),
        )
        self.assertEqual(
            'https://sp2.com/v3/OS-FEDERATION/'
            'identity_providers/acme/protocols/saml2/auth',
            service_providers.get_auth_url('sp2'),
        )
        self.assertEqual(
            'https://sp2.com/Shibboleth.sso/SAML2/ECP',
            service_providers.get_sp_url('sp2'),
        )

    def test_handle_missing_service_provider(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)

        a = v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session()
        auth_ref = a.get_auth_ref(s)

        service_providers = auth_ref.service_providers

        self.assertRaises(
            exceptions.ServiceProviderNotFound,
            service_providers._get_service_provider,
            uuid.uuid4().hex,
        )

    def test_invalid_auth_response_dict(self):
        self.stub_auth(json={'hello': 'world'})

        a = v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        self.assertRaises(
            exceptions.InvalidResponse, s.get, 'http://any', authenticated=True
        )

    def test_invalid_auth_response_type(self):
        self.stub_url('POST', ['auth', 'tokens'], text='testdata')

        a = v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        self.assertRaises(
            exceptions.InvalidResponse, s.get, 'http://any', authenticated=True
        )

    def test_invalidate_response(self):
        auth_responses = [
            {
                'status_code': 200,
                'json': self.TEST_RESPONSE_DICT,
                'headers': {'X-Subject-Token': 'token1'},
            },
            {
                'status_code': 200,
                'json': self.TEST_RESPONSE_DICT,
                'headers': {'X-Subject-Token': 'token2'},
            },
        ]

        self.requests_mock.post(f'{self.TEST_URL}/auth/tokens', auth_responses)

        a = v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        self.assertEqual('token1', s.get_token())
        self.assertEqual({'X-Auth-Token': 'token1'}, s.get_auth_headers())
        a.invalidate()
        self.assertEqual('token2', s.get_token())
        self.assertEqual({'X-Auth-Token': 'token2'}, s.get_auth_headers())

    def test_doesnt_log_password(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)

        password = uuid.uuid4().hex
        a = v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=password
        )
        s = session.Session(a)
        self.assertEqual(self.TEST_TOKEN, s.get_token())
        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        self.assertNotIn(password, self.logger.output)

    def test_sends_nocatalog(self):
        del self.TEST_RESPONSE_DICT['token']['catalog']
        self.stub_auth(json=self.TEST_RESPONSE_DICT)

        a = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            include_catalog=False,
        )
        s = session.Session(auth=a)

        s.get_token()

        auth_url = self.TEST_URL + '/auth/tokens'
        self.assertEqual(auth_url, a.token_url)
        self.assertEqual(
            auth_url + '?nocatalog', self.requests_mock.last_request.url
        )

    def test_symbols(self):
        self.assertIs(v3.AuthMethod, v3_base.AuthMethod)
        self.assertIs(v3.AuthConstructor, v3_base.AuthConstructor)
        self.assertIs(v3.Auth, v3_base.Auth)

    def test_unscoped_request(self):
        token = fixture.V3Token()
        self.stub_auth(json=token)
        password = uuid.uuid4().hex

        a = v3.Password(
            self.TEST_URL,
            user_id=token.user_id,
            password=password,
            unscoped=True,
        )
        s = session.Session()

        auth_ref = a.get_access(s)

        self.assertFalse(auth_ref.scoped)
        body = self.requests_mock.last_request.json()

        ident = body['auth']['identity']

        self.assertEqual(['password'], ident['methods'])
        self.assertEqual(token.user_id, ident['password']['user']['id'])
        self.assertEqual(password, ident['password']['user']['password'])

        self.assertEqual('unscoped', body['auth']['scope'])

    def test_unscoped_with_scope_data(self):
        a = v3.Password(
            self.TEST_URL,
            user_id=uuid.uuid4().hex,
            password=uuid.uuid4().hex,
            unscoped=True,
            project_id=uuid.uuid4().hex,
        )

        s = session.Session()

        self.assertRaises(exceptions.AuthorizationFailure, a.get_auth_ref, s)

    def test_password_cache_id(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        project_name = uuid.uuid4().hex

        a = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            user_domain_id=self.TEST_DOMAIN_ID,
            project_domain_name=self.TEST_DOMAIN_NAME,
            project_name=project_name,
        )

        b = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            user_domain_id=self.TEST_DOMAIN_ID,
            project_domain_name=self.TEST_DOMAIN_NAME,
            project_name=project_name,
        )

        a_id = a.get_cache_id()
        b_id = b.get_cache_id()

        self.assertEqual(a_id, b_id)

        c = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            user_domain_id=self.TEST_DOMAIN_ID,
            project_domain_name=self.TEST_DOMAIN_NAME,
            project_id=project_name,
        )  # same value different param

        c_id = c.get_cache_id()

        self.assertNotEqual(a_id, c_id)

        self.assertIsNone(a.get_auth_state())
        self.assertIsNone(b.get_auth_state())
        self.assertIsNone(c.get_auth_state())

        s = session.Session()
        self.assertEqual(self.TEST_TOKEN, a.get_token(s))
        self.assertTrue(self.requests_mock.called)

    def test_password_change_auth_state(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)

        expired = ksa_utils.before_utcnow(days=2)
        token = fixture.V3Token(expires=expired)
        token_id = uuid.uuid4().hex

        state = json.dumps({'auth_token': token_id, 'body': token})

        a = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            user_domain_id=self.TEST_DOMAIN_ID,
            project_id=uuid.uuid4().hex,
        )

        initial_cache_id = a.get_cache_id()

        self.assertIsNone(a.get_auth_state())
        a.set_auth_state(state)

        self.assertEqual(token_id, a.auth_ref.auth_token)

        s = session.Session()
        self.assertEqual(self.TEST_TOKEN, a.get_token(s))  # updates expired
        self.assertEqual(initial_cache_id, a.get_cache_id())

    def test_receipt_response_is_handled(self):
        self.stub_receipt()

        a = v3.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            user_domain_id=self.TEST_DOMAIN_ID,
            project_id=self.TEST_TENANT_ID,
        )

        s = session.Session(a)
        self.assertRaises(
            exceptions.MissingAuthMethods, s.get_auth_headers, None
        )

    def test_authenticate_with_receipt_and_totp(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        passcode = "123456"
        auth = v3.TOTP(
            self.TEST_URL, username=self.TEST_USER, passcode=passcode
        )
        auth.add_method(v3.ReceiptMethod(receipt=self.TEST_RECEIPT))
        self.assertFalse(auth.has_scope_parameters)
        s = session.Session(auth=auth)

        self.assertEqual(
            {"X-Auth-Token": self.TEST_TOKEN}, s.get_auth_headers()
        )

        # NOTE(adriant): Here we are confirming the receipt data isn't in the
        # body or listed as a method
        req = {
            "auth": {
                "identity": {
                    "methods": ["totp"],
                    "totp": {
                        "user": {"name": self.TEST_USER, "passcode": passcode}
                    },
                }
            }
        }

        self.assertRequestBodyIs(json=req)
        self.assertRequestHeaderEqual(
            "Openstack-Auth-Receipt", self.TEST_RECEIPT
        )
        self.assertRequestHeaderEqual("Content-Type", "application/json")
        self.assertRequestHeaderEqual("Accept", "application/json")
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_authenticate_with_multi_factor(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        passcode = "123456"
        auth = v3.MultiFactor(
            self.TEST_URL,
            auth_methods=['v3password', 'v3totp'],
            username=self.TEST_USER,
            password=self.TEST_PASS,
            passcode=passcode,
            user_domain_id=self.TEST_DOMAIN_ID,
            project_id=self.TEST_TENANT_ID,
        )
        self.assertTrue(auth.has_scope_parameters)
        s = session.Session(auth=auth)

        self.assertEqual(
            {"X-Auth-Token": self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            "auth": {
                "identity": {
                    "methods": ["password", "totp"],
                    "totp": {
                        "user": {
                            "name": self.TEST_USER,
                            "passcode": passcode,
                            'domain': {'id': self.TEST_DOMAIN_ID},
                        }
                    },
                    'password': {
                        'user': {
                            'name': self.TEST_USER,
                            'password': self.TEST_PASS,
                            'domain': {'id': self.TEST_DOMAIN_ID},
                        }
                    },
                },
                'scope': {'project': {'id': self.TEST_TENANT_ID}},
            }
        }

        self.assertRequestBodyIs(json=req)
        self.assertRequestHeaderEqual("Content-Type", "application/json")
        self.assertRequestHeaderEqual("Accept", "application/json")
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_authenticate_with_unversioned_endpoint(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        # We use the root url here because it doesn't reference the API version
        # (e.g., '/v3'). We want to make sure the authentication plugin handles
        # this and appends /v3 if it's not present.
        a = v3.Password(
            self.TEST_ROOT_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
        )
        self.assertFalse(a.has_scope_parameters)
        s = session.Session(auth=a)

        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'identity': {
                    'methods': ['password'],
                    'password': {
                        'user': {
                            'name': self.TEST_USER,
                            'password': self.TEST_PASS,
                        }
                    },
                }
            }
        }

        self.assertRequestBodyIs(json=req)
        self.assertRequestHeaderEqual('Content-Type', 'application/json')
        self.assertRequestHeaderEqual('Accept', 'application/json')
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_oauth2_client_credential_method_http(self):
        base_http = self.TEST_URL
        oauth2_endpoint = f'{self.TEST_URL}/oauth_token'
        oauth2_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        self.stub_auth(json=self.TEST_APP_CRED_TOKEN_RESPONSE)
        client_cre = v3.OAuth2ClientCredential(
            base_http,
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=self.TEST_CLIENT_CRED_ID,
            oauth2_client_secret=self.TEST_CLIENT_CRED_SECRET,
        )
        oauth2_resp = {
            'status_code': 200,
            'json': {
                'access_token': oauth2_token,
                'expires_in': 3600,
                'token_type': 'Bearer',
            },
        }
        self.requests_mock.post(oauth2_endpoint, [oauth2_resp])

        sess = session.Session(auth=client_cre)
        initial_cache_id = client_cre.get_cache_id()

        auth_head = sess.get_auth_headers()
        self.assertEqual(self.TEST_TOKEN, auth_head['X-Auth-Token'])
        self.assertEqual(f'Bearer {oauth2_token}', auth_head['Authorization'])

        self.assertEqual(sess.auth.auth_ref.auth_token, self.TEST_TOKEN)
        self.assertEqual(initial_cache_id, client_cre.get_cache_id())

        resp_ok = {'status_code': 200}
        self.requests_mock.post(f'{base_http}/test_api', [resp_ok])
        resp = sess.post(f'{base_http}/test_api', authenticated=True)
        self.assertRequestHeaderEqual(
            'Authorization', f'Bearer {oauth2_token}'
        )
        self.assertRequestHeaderEqual('X-Auth-Token', self.TEST_TOKEN)
        self.assertEqual(200, resp.status_code)

    def test_oauth2_client_credential_method_https(self):
        self.TEST_URL = self.TEST_URL.replace('http:', 'https:')
        base_https = self.TEST_URL
        oauth2_endpoint = f'{base_https}/oauth_token'
        oauth2_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        self.stub_auth(json=self.TEST_APP_CRED_TOKEN_RESPONSE)
        client_cre = v3.OAuth2ClientCredential(
            base_https,
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=self.TEST_CLIENT_CRED_ID,
            oauth2_client_secret=self.TEST_CLIENT_CRED_SECRET,
        )
        oauth2_resp = {
            'status_code': 200,
            'json': {
                'access_token': oauth2_token,
                'expires_in': 3600,
                'token_type': 'Bearer',
            },
        }
        self.requests_mock.post(oauth2_endpoint, [oauth2_resp])

        sess = session.Session(auth=client_cre)
        initial_cache_id = client_cre.get_cache_id()

        auth_head = sess.get_auth_headers()
        self.assertEqual(self.TEST_TOKEN, auth_head['X-Auth-Token'])
        self.assertEqual(f'Bearer {oauth2_token}', auth_head['Authorization'])

        self.assertEqual(sess.auth.auth_ref.auth_token, self.TEST_TOKEN)
        self.assertEqual(initial_cache_id, client_cre.get_cache_id())

        resp_ok = {'status_code': 200}
        self.requests_mock.post(f'{base_https}/test_api', [resp_ok])
        resp = sess.post(f'{base_https}/test_api', authenticated=True)
        self.assertRequestHeaderEqual(
            'Authorization', f'Bearer {oauth2_token}'
        )
        self.assertRequestHeaderEqual('X-Auth-Token', self.TEST_TOKEN)
        self.assertEqual(200, resp.status_code)

    def test_oauth2_client_credential_method_base_header_none(self):
        base_https = self.TEST_URL.replace('http:', 'https:')
        oauth2_endpoint = f'{base_https}/oauth_token'
        oauth2_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        with unittest.mock.patch(
            'keystoneauth1.plugin.BaseAuthPlugin.get_headers'
        ) as co_mock:
            co_mock.return_value = None
            client_cre = v3.OAuth2ClientCredential(
                base_https,
                oauth2_endpoint=oauth2_endpoint,
                oauth2_client_id=self.TEST_CLIENT_CRED_ID,
                oauth2_client_secret=self.TEST_CLIENT_CRED_SECRET,
            )
            oauth2_resp = {
                'status_code': 200,
                'json': {
                    'access_token': oauth2_token,
                    'expires_in': 3600,
                    'token_type': 'Bearer',
                },
            }
            self.requests_mock.post(oauth2_endpoint, [oauth2_resp])

            sess = session.Session(auth=client_cre)
            auth_head = sess.get_auth_headers()
            self.assertNotIn('X-Auth-Token', auth_head)
            self.assertEqual(
                f'Bearer {oauth2_token}', auth_head['Authorization']
            )

    def test_oauth2_client_credential_method_rm_auth(self):
        base_https = self.TEST_URL.replace('http:', 'https:')
        base_http = self.TEST_URL
        oauth2_endpoint = f'{base_https}/oauth_token'
        oauth2_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        self.stub_auth(json=self.TEST_APP_CRED_TOKEN_RESPONSE)
        client_cre = v3.OAuth2ClientCredential(
            base_http,
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=self.TEST_CLIENT_CRED_ID,
            oauth2_client_secret=self.TEST_CLIENT_CRED_SECRET,
        )
        oauth2_resp = {
            'status_code': 200,
            'json': {
                'access_token': oauth2_token,
                'expires_in': 3600,
                'token_type': 'Bearer',
            },
        }
        self.requests_mock.post(oauth2_endpoint, [oauth2_resp])

        sess = session.Session(auth=client_cre)
        initial_cache_id = client_cre.get_cache_id()

        auth_head = sess.get_auth_headers()
        self.assertEqual(self.TEST_TOKEN, auth_head['X-Auth-Token'])
        self.assertEqual(f'Bearer {oauth2_token}', auth_head['Authorization'])

        self.assertEqual(sess.auth.auth_ref.auth_token, self.TEST_TOKEN)
        self.assertEqual(initial_cache_id, client_cre.get_cache_id())

        resp_ok = {'status_code': 200}
        self.requests_mock.post(f'{base_http}/test_api', [resp_ok])
        resp = sess.post(f'{base_http}/test_api', authenticated=True)
        self.assertRequestHeaderEqual(
            'Authorization', f'Bearer {oauth2_token}'
        )
        self.assertRequestHeaderEqual('X-Auth-Token', self.TEST_TOKEN)
        self.assertEqual(200, resp.status_code)

    def test_oauth2_client_credential_method_other_not_rm_auth(self):
        base_https = self.TEST_URL.replace('http:', 'https:')
        other_auth_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        self.stub_auth(json=self.TEST_APP_CRED_TOKEN_RESPONSE)
        with unittest.mock.patch(
            'keystoneauth1.identity.v3.Password.get_headers'
        ) as co_mock:
            co_mock.return_value = {
                'X-Auth-Token': self.TEST_TOKEN,
                'Authorization': other_auth_token,
            }
            pass_auth = v3.Password(
                base_https,
                username=self.TEST_USER,
                password=self.TEST_PASS,
                include_catalog=False,
            )
            sess = session.Session(auth=pass_auth)

            resp_ok = {'status_code': 200}
            self.requests_mock.post(f'{base_https}/test_api', [resp_ok])
            resp = sess.post(f'{base_https}/test_api', authenticated=True)
            self.assertRequestHeaderEqual('Authorization', other_auth_token)
            self.assertRequestHeaderEqual('X-Auth-Token', self.TEST_TOKEN)
            self.assertEqual(200, resp.status_code)

    def test_oauth2_client_credential_method_500(self):
        self.TEST_URL = self.TEST_URL.replace('http:', 'https:')
        base_https = self.TEST_URL
        oauth2_endpoint = f'{base_https}/oauth_token'
        self.stub_auth(json=self.TEST_APP_CRED_TOKEN_RESPONSE)
        client_cre = v3.OAuth2ClientCredential(
            base_https,
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=self.TEST_CLIENT_CRED_ID,
            oauth2_client_secret=self.TEST_CLIENT_CRED_SECRET,
        )
        oauth2_resp = {
            'status_code': 500,
            'json': {
                'error': 'other_error',
                'error_description': 'Unknown error is occur.',
            },
        }
        self.requests_mock.post(oauth2_endpoint, [oauth2_resp])

        sess = session.Session(auth=client_cre)
        err = self.assertRaises(ClientException, sess.get_auth_headers)
        self.assertEqual('Unknown error is occur.', str(err))

    def test_oauth2_client_credential_reauth_called_https(self):
        base_https = self.TEST_URL.replace('http:', 'https:')
        oauth2_endpoint = f'{base_https}/oauth_token'
        oauth2_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        auth = v3.OAuth2ClientCredential(
            base_https,
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id='clientcredid',
            oauth2_client_secret='secret',
        )
        oauth2_resp = {
            'status_code': 200,
            'json': {
                'access_token': oauth2_token,
                'expires_in': 3600,
                'token_type': 'Bearer',
            },
        }
        self.requests_mock.post(oauth2_endpoint, [oauth2_resp])

        sess = session.Session(auth=auth)

        resp_text = json.dumps(self.TEST_APP_CRED_TOKEN_RESPONSE)
        resp_ok = {
            'status_code': 200,
            'headers': {
                'Content-Type': 'application/json',
                'x-subject-token': self.TEST_TOKEN,
            },
            'text': resp_text,
        }
        self.requests_mock.post(
            f'{base_https}/auth/tokens',
            [resp_ok, {'text': 'Failed', 'status_code': 401}, resp_ok],
        )

        resp = sess.post(f'{base_https}/auth/tokens', authenticated=True)
        self.assertRequestHeaderEqual(
            'Authorization', f'Bearer {oauth2_token}'
        )
        self.assertEqual(200, resp.status_code)
        self.assertEqual(resp_text, resp.text)

    def test_oauth2_mtls_client_credential_method(self):
        base_https = self.TEST_URL.replace('http:', 'https:')
        token_endpoint = f'{self.TEST_URL}/auth/tokens'
        oauth2_endpoint = f'{base_https}/OS-OAUTH2/token'
        oauth2_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        a = v3.OAuth2mTlsClientCredential(
            self.TEST_URL,
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=self.TEST_CLIENT_CRED_ID,
        )

        oauth2_post_resp = {
            'status_code': 200,
            'json': {
                'access_token': oauth2_token,
                'expires_in': 3600,
                'token_type': 'Bearer',
            },
        }
        self.requests_mock.post(oauth2_endpoint, [oauth2_post_resp])
        token_verify_resp = {
            'status_code': 200,
            'json': {**self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE},
        }
        self.requests_mock.get(token_endpoint, [token_verify_resp])

        sess = session.Session(auth=a)
        auth_ref = a.get_auth_ref(sess)
        self.assertEqual(auth_ref.auth_token, oauth2_token)
        self.assertEqual(auth_ref._data, self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE)
        self.assertEqual(
            auth_ref.project_id,
            self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE.get('token', {})
            .get('project', {})
            .get('id'),
        )
        self.assertIsNone(auth_ref.domain_id)
        self.assertEqual(
            auth_ref.oauth2_credential,
            self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE.get('token', {}).get(
                'oauth2_credential'
            ),
        )
        self.assertEqual(
            auth_ref.oauth2_credential_thumbprint,
            self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE.get('token', {})
            .get('oauth2_credential', {})
            .get('x5t#S256'),
        )

        auth_head = sess.get_auth_headers()
        self.assertEqual(f'Bearer {oauth2_token}', auth_head['Authorization'])
        self.assertEqual(oauth2_token, auth_head['X-Auth-Token'])

    def test_oauth2_mtls_client_credential_method_without_v3(self):
        base_https = self.TEST_URL.replace('http:', 'https:')
        token_endpoint = f'{self.TEST_URL}/auth/tokens'
        oauth2_endpoint = f'{base_https}/OS-OAUTH2/token'
        oauth2_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        a = v3.OAuth2mTlsClientCredential(
            self.TEST_URL.replace('v3', ''),
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=self.TEST_CLIENT_CRED_ID,
        )

        oauth2_post_resp = {
            'status_code': 200,
            'json': {
                'access_token': oauth2_token,
                'expires_in': 3600,
                'token_type': 'Bearer',
            },
        }
        self.requests_mock.post(oauth2_endpoint, [oauth2_post_resp])
        token_verify_resp = {
            'status_code': 200,
            'json': {**self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE},
        }
        self.requests_mock.get(token_endpoint, [token_verify_resp])

        sess = session.Session(auth=a)
        auth_ref = a.get_auth_ref(sess)
        self.assertEqual(auth_ref.auth_token, oauth2_token)
        self.assertEqual(auth_ref._data, self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE)
        self.assertEqual(
            auth_ref.oauth2_credential,
            self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE.get('token', {}).get(
                'oauth2_credential'
            ),
        )
        self.assertEqual(
            auth_ref.oauth2_credential_thumbprint,
            self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE.get('token', {})
            .get('oauth2_credential', {})
            .get('x5t#S256'),
        )
        auth_head = sess.get_auth_headers()
        self.assertEqual(f'Bearer {oauth2_token}', auth_head['Authorization'])
        self.assertEqual(oauth2_token, auth_head['X-Auth-Token'])

    def test_oauth2_mtls_client_credential_method_resp_invalid_json(self):
        base_https = self.TEST_URL.replace('http:', 'https:')
        token_endpoint = f'{self.TEST_URL}/auth/tokens'
        oauth2_endpoint = f'{base_https}/OS-OAUTH2/token'
        oauth2_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        a = v3.OAuth2mTlsClientCredential(
            self.TEST_URL,
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=self.TEST_CLIENT_CRED_ID,
        )

        oauth2_post_resp = {
            'status_code': 200,
            'json': {
                'access_token': oauth2_token,
                'expires_in': 3600,
                'token_type': 'Bearer',
            },
        }
        self.requests_mock.post(oauth2_endpoint, [oauth2_post_resp])
        token_verify_resp = {'status_code': 200, 'text': 'invalid json'}
        self.requests_mock.get(token_endpoint, [token_verify_resp])

        sess = session.Session(auth=a)
        self.assertRaises(exceptions.InvalidResponse, a.get_auth_ref, sess)

    def test_oauth2_mtls_client_credential_method_resp_without_token(self):
        base_https = self.TEST_URL.replace('http:', 'https:')
        token_endpoint = f'{self.TEST_URL}/auth/tokens'
        oauth2_endpoint = f'{base_https}/OS-OAUTH2/token'
        oauth2_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        a = v3.OAuth2mTlsClientCredential(
            self.TEST_URL,
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=self.TEST_CLIENT_CRED_ID,
        )

        oauth2_post_resp = {
            'status_code': 200,
            'json': {
                'access_token': oauth2_token,
                'expires_in': 3600,
                'token_type': 'Bearer',
            },
        }
        self.requests_mock.post(oauth2_endpoint, [oauth2_post_resp])
        token_verify_resp = {'status_code': 200, 'json': {'without_token': {}}}
        self.requests_mock.get(token_endpoint, [token_verify_resp])

        sess = session.Session(auth=a)
        self.assertRaises(exceptions.InvalidResponse, a.get_auth_ref, sess)

    def test_oauth2_mtls_client_credential_method_client_exception(self):
        base_https = self.TEST_URL.replace('http:', 'https:')
        oauth2_endpoint = f'{base_https}/OS-OAUTH2/token'
        a = v3.OAuth2mTlsClientCredential(
            self.TEST_URL,
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=self.TEST_CLIENT_CRED_ID,
        )

        oauth2_post_resp = {'status_code': 400, 'json': {}}
        self.requests_mock.post(oauth2_endpoint, [oauth2_post_resp])

        sess = session.Session(auth=a)
        self.assertRaises(exceptions.ClientException, a.get_auth_ref, sess)

    def test_oauth2_mtls_client_credential_method_base_header_none(self):
        base_https = self.TEST_URL.replace('http:', 'https:')
        token_endpoint = f'{self.TEST_URL}/auth/tokens'
        oauth2_endpoint = f'{base_https}/OS-OAUTH2/token'
        oauth2_token = 'HW9bB6oYWJywz6mAN_KyIBXlof15Pk'
        a = v3.OAuth2mTlsClientCredential(
            self.TEST_URL,
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=self.TEST_CLIENT_CRED_ID,
        )
        oauth2_post_resp = {
            'status_code': 200,
            'json': {
                'access_token': oauth2_token,
                'expires_in': 3600,
                'token_type': 'Bearer',
            },
        }
        self.requests_mock.post(oauth2_endpoint, [oauth2_post_resp])
        token_verify_resp = {
            'status_code': 200,
            'json': {**self.TEST_OAUTH2_MTLS_TOKEN_RESPONSE},
        }
        self.requests_mock.get(token_endpoint, [token_verify_resp])
        sess = session.Session(auth=a)

        with unittest.mock.patch(
            'keystoneauth1.plugin.BaseAuthPlugin.get_headers'
        ) as co_mock:
            co_mock.return_value = None
            auth_head = sess.get_auth_headers()
            self.assertEqual('Bearer None', auth_head['Authorization'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import json
import typing as ty
import warnings

import typing_extensions as ty_ext

from keystoneauth1 import _utils as utils
from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1.identity import base
from keystoneauth1 import session as ks_session

LOG = utils.get_logger(__name__)

__all__ = ('Auth', 'AuthMethod', 'AuthConstructor', 'BaseAuth')


class BaseAuth(base.BaseIdentityPlugin, metaclass=abc.ABCMeta):
    """Identity V3 Authentication Plugin.

    :param string auth_url: Identity service endpoint for authentication.
    :param string trust_id: Trust ID for trust scoping.
    :param string system_scope: System information to scope to.
    :param string domain_id: Domain ID for domain scoping.
    :param string domain_name: Domain name for domain scoping.
    :param string project_id: Project ID for project scoping.
    :param string project_name: Project name for project scoping.
    :param string project_domain_id: Project's domain ID for project.
    :param string project_domain_name: Project's domain name for project.
    :param bool reauthenticate: Allow fetching a new token if the current one
                                is going to expire. (optional) default True
    :param bool include_catalog: Include the service catalog in the returned
                                 token. (optional) default True.
    """

    auth_url: str

    def __init__(
        self,
        auth_url: str,
        *,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        super().__init__(auth_url=auth_url, reauthenticate=reauthenticate)

        self.trust_id = trust_id
        self.system_scope = system_scope
        self.domain_id = domain_id
        self.domain_name = domain_name
        self.project_id = project_id
        self.project_name = project_name
        self.project_domain_id = project_domain_id
        self.project_domain_name = project_domain_name
        self.include_catalog = include_catalog

    @property
    def token_url(self) -> str:
        """The full URL where we will send authentication data."""
        return '{}/auth/tokens'.format(self.auth_url.rstrip('/'))

    @property
    def has_scope_parameters(self) -> bool:
        """Return true if parameters can be used to create a scoped token."""
        return bool(
            self.domain_id
            or self.domain_name
            or self.project_id
            or self.project_name
            or self.trust_id
            or self.system_scope
        )


class _AuthIdentity(ty.TypedDict):
    identity: dict[str, ty.Any]
    scope: ty_ext.NotRequired[dict[str, ty.Any] | str]


class _AuthBody(ty.TypedDict):
    auth: _AuthIdentity


class Auth(BaseAuth):
    """Identity V3 Authentication Plugin.

    :param string auth_url: Identity service endpoint for authentication.
    :param list auth_methods: A collection of methods to authenticate with.
    :param string trust_id: Trust ID for trust scoping.
    :param string domain_id: Domain ID for domain scoping.
    :param string domain_name: Domain name for domain scoping.
    :param string project_id: Project ID for project scoping.
    :param string project_name: Project name for project scoping.
    :param string project_domain_id: Project's domain ID for project.
    :param string project_domain_name: Project's domain name for project.
    :param bool reauthenticate: Allow fetching a new token if the current one
                                is going to expire. (optional) default True
    :param bool include_catalog: Include the service catalog in the returned
                                 token. (optional) default True.
    :param bool unscoped: Force the return of an unscoped token. This will make
                          the keystone server return an unscoped token even if
                          a default_project_id is set for this user.
    """

    def __init__(
        self,
        auth_url: str,
        auth_methods: list['AuthMethod'],
        *,
        unscoped: bool = False,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        super().__init__(
            auth_url=auth_url,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
        self.unscoped = unscoped
        self.auth_methods = auth_methods

    def add_method(self, method: 'AuthMethod') -> None:
        """Add an additional initialized AuthMethod instance."""
        self.auth_methods.append(method)

    def get_auth_ref(self, session: ks_session.Session) -> access.AccessInfoV3:
        headers = {'Accept': 'application/json'}
        body: _AuthBody = {'auth': {'identity': {}}}
        ident = body['auth']['identity']
        # this is passed around for its side-effects
        rkwargs: dict[str, ty.Any] = {}

        for method in self.auth_methods:
            name, auth_data = method.get_auth_data(
                session, self, headers, request_kwargs=rkwargs
            )
            # NOTE(adriant): Methods like ReceiptMethod don't
            # want anything added to the request data, so they
            # explicitly return None, which we check for.
            if name:
                ident.setdefault('methods', []).append(name)
                ident[name] = auth_data

        if not ident:
            raise exceptions.AuthorizationFailure(
                'Authentication method required (e.g. password)'
            )

        mutual_exclusion = [
            bool(self.domain_id or self.domain_name),
            bool(self.project_id or self.project_name),
            bool(self.trust_id),
            bool(self.system_scope),
            bool(self.unscoped),
        ]

        if sum(mutual_exclusion) > 1:
            raise exceptions.AuthorizationFailure(
                message='Authentication cannot be scoped to multiple'
                ' targets. Pick one of: project, domain, '
                'trust, system or unscoped'
            )

        if self.domain_id:
            body['auth']['scope'] = {'domain': {'id': self.domain_id}}
        elif self.domain_name:
            body['auth']['scope'] = {'domain': {'name': self.domain_name}}
        elif self.project_id:
            body['auth']['scope'] = {'project': {'id': self.project_id}}
        elif self.project_name:
            scope = body['auth']['scope'] = {'project': {}}
            scope['project']['name'] = self.project_name

            if self.project_domain_id:
                scope['project']['domain'] = {'id': self.project_domain_id}
            elif self.project_domain_name:
                scope['project']['domain'] = {'name': self.project_domain_name}
        elif self.trust_id:
            body['auth']['scope'] = {'OS-TRUST:trust': {'id': self.trust_id}}
        elif self.unscoped:
            body['auth']['scope'] = 'unscoped'
        elif self.system_scope:
            # NOTE(lbragstad): Right now it's only possible to have role
            # assignments on the entire system. In the future that might change
            # so that users and groups can have roles on parts of the system,
            # like a specific service in a specific region. If that happens,
            # this will have to be accounted for here. Until then we'll only
            # support scoping to the entire system.
            if self.system_scope == 'all':
                body['auth']['scope'] = {'system': {'all': True}}

        token_url = self.token_url

        if not self.auth_url.rstrip('/').endswith('v3'):
            token_url = '{}/v3/auth/tokens'.format(self.auth_url.rstrip('/'))

        # NOTE(jamielennox): we add nocatalog here rather than in token_url
        # directly as some federation plugins require the base token_url
        if not self.include_catalog:
            token_url += '?nocatalog'

        LOG.debug('Making authentication request to %s', token_url)
        resp = session.post(
            token_url,
            json=body,
            headers=headers,
            authenticated=False,
            log=False,
            **rkwargs,
        )

        try:
            LOG.debug(json.dumps(resp.json()))
            resp_data = resp.json()
        except ValueError:
            raise exceptions.InvalidResponse(response=resp)

        if 'token' not in resp_data:
            raise exceptions.InvalidResponse(response=resp)

        return access.AccessInfoV3(
            auth_token=resp.headers['X-Subject-Token'], body=resp_data
        )

    def get_cache_id_elements(self) -> dict[str, str | None]:
        if not self.auth_methods:
            return {}

        params = {
            'auth_url': self.auth_url,
            'domain_id': self.domain_id,
            'domain_name': self.domain_name,
            'project_id': self.project_id,
            'project_name': self.project_name,
            'project_domain_id': self.project_domain_id,
            'project_domain_name': self.project_domain_name,
            'trust_id': self.trust_id,
        }

        for method in self.auth_methods:
            # may raise NotImplementedError but that's handled by
            # BaseIdentityPlugin.get_cache_id
            elements = method.get_cache_id_elements()
            params.update(elements)

        return params


class AuthMethod(metaclass=abc.ABCMeta):
    """One part of a V3 Authentication strategy.

    The v3 '/tokens' API allow multiple methods to be presented when
    authentication against the server. Each one of these methods is implemented
    by an AuthMethod.

    Note: When implementing an AuthMethod use keyword arguments to ensure they
    are supported by the MultiFactor auth plugin.
    """

    #: Deprecated parameter for defining the parameters supported by the
    #: plugin. These should now be defined by typed class attributes.
    _method_parameters: list[str] | None = None

    # TODO(stephenfin): Remove support for arbitrary arguments in 2025.2 or
    # later
    def __init__(self, **kwargs: object):
        if self._method_parameters is not None:
            warnings.warn(
                "Defining method parameter via '_method_parameters' is "
                "deprecated and will be removed in a future release. Migrate "
                "to typed class attributes and define an '__init__' method.",
                category=DeprecationWarning,
            )

            for param in self._method_parameters:
                setattr(self, param, kwargs.pop(param, None))

        if kwargs:
            msg = "Unexpected Attributes: {}".format(", ".join(kwargs.keys()))
            raise AttributeError(msg)

    @classmethod
    def _extract_kwargs(cls, kwargs: dict[str, object]) -> dict[str, object]:
        """Remove parameters related to this method from other kwargs."""
        _method_parameters = cls._method_parameters or []
        return {p: kwargs.pop(p, None) for p in _method_parameters}

    @abc.abstractmethod
    def get_auth_data(
        self,
        session: ks_session.Session,
        auth: Auth,
        headers: dict[str, str],
        request_kwargs: dict[str, object],
    ) -> tuple[None, None] | tuple[str, ty.Mapping[str, object]]:
        """Return the authentication section of an auth plugin.

        :param session: The communication session.
        :type session: keystoneauth1.session.Session
        :param base.Auth auth: The auth plugin calling the method.
        :param dict headers: The headers that will be sent with the auth
                             request if a plugin needs to add to them.
        :return: The identifier of this plugin and a dict of authentication
                 data for the auth type.
        :rtype: tuple(string, dict)
        """
        raise NotImplementedError()

    def get_cache_id_elements(self) -> dict[str, str | None]:
        """Get the elements for this auth method that make it unique.

        These elements will be used as part of the
        :py:meth:`keystoneauth1.plugin.BaseIdentityPlugin.get_cache_id` to
        allow caching of the auth plugin.

        Plugins should override this if they want to allow caching of their
        state.

        To avoid collision or overrides the keys of the returned dictionary
        should be prefixed with the plugin identifier. For example the password
        plugin returns its username value as 'password_username'.
        """
        raise NotImplementedError()


@ty.runtime_checkable
class SupportsMultiFactor(ty.Protocol):
    _auth_method_class: ty.ClassVar[type[AuthMethod]]


class AuthConstructor(Auth, metaclass=abc.ABCMeta):
    """Abstract base class for creating an Auth Plugin.

    The Auth Plugin created contains only one authentication method. This
    is generally the required usage.

    An AuthConstructor creates an AuthMethod based on the method's
    arguments and the auth_method_class defined by the plugin. It then
    creates the auth plugin with only that authentication method.
    """

    _auth_method_class: ty.ClassVar[type[AuthMethod]]

    def __init__(
        self,
        auth_url: str,
        *args: ty.Any,
        unscoped: bool = False,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
        **kwargs: ty.Any,
    ):
        warnings.warn(
            f"'{AuthConstructor.__qualname__}' is deprecated and will be "
            f"removed in a future release. Subclass '{Auth.__qualname__}' "
            f"instead and define typed class attributes and an '__init__' "
            f"method.",
            category=DeprecationWarning,
        )

        method_kwargs = self._auth_method_class._extract_kwargs(kwargs)
        # we should have consumed all "unknown" arguments by now
        assert kwargs == {}  # nosec B101
        method = self._auth_method_class(*args, **method_kwargs)
        super().__init__(
            auth_url,
            [method],
            unscoped=unscoped,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from keystoneauth1.extras.oauth1 import v3
from keystoneauth1 import loading
from keystoneauth1.loading import opts


# NOTE(jamielennox): This is not a BaseV3Loader because we don't want to
# include the scoping options like project-id in the option list
class V3OAuth1(loading.BaseIdentityLoader[v3.OAuth1]):
    @property
    def plugin_class(self) -> type[v3.OAuth1]:
        return v3.OAuth1

    @property
    def available(self) -> bool:
        return v3.oauth1 is not None

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'consumer-key', required=True, help='OAuth Consumer ID/Key'
                ),
                loading.Opt(
                    'consumer-secret',
                    required=True,
                    help='OAuth Consumer Secret',
                ),
                loading.Opt(
                    'access-key', required=True, help='OAuth Access Key'
                ),
                loading.Opt(
                    'access-secret', required=True, help='OAuth Access Secret'
                ),
            ]
        )

        return options



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1 import fixture
from keystoneauth1 import session
from keystoneauth1.tests.unit import utils


class GenericPluginTestCase(utils.TestCase):
    TEST_URL = 'http://keystone.host:5000/'

    # OVERRIDE THESE IN SUB CLASSES
    PLUGIN_CLASS = None
    V2_PLUGIN_CLASS = None
    V3_PLUGIN_CLASS = None

    def setUp(self):
        super().setUp()

        self.token_v2 = fixture.V2Token()
        self.token_v3 = fixture.V3Token()
        self.token_v3_id = uuid.uuid4().hex
        self.session = session.Session()

        self.stub_url('POST', ['v2.0', 'tokens'], json=self.token_v2)
        self.stub_url(
            'POST',
            ['v3', 'auth', 'tokens'],
            headers={'X-Subject-Token': self.token_v3_id},
            json=self.token_v3,
        )

    def new_plugin(self, **kwargs):
        kwargs.setdefault('auth_url', self.TEST_URL)
        return self.PLUGIN_CLASS(**kwargs)

    def stub_discovery(self, base_url=None, **kwargs):
        kwargs.setdefault('href', self.TEST_URL)
        disc = fixture.DiscoveryList(**kwargs)
        self.stub_url('GET', json=disc, base_url=base_url, status_code=300)
        return disc

    def assertCreateV3(self, **kwargs):
        auth = self.new_plugin(**kwargs)
        auth_ref = auth.get_auth_ref(self.session)
        self.assertIsInstance(auth_ref, access.AccessInfoV3)
        self.assertEqual(
            self.TEST_URL + 'v3/auth/tokens',
            self.requests_mock.last_request.url,
        )
        self.assertIsInstance(auth._plugin, self.V3_PLUGIN_CLASS)
        return auth

    def assertCreateV2(self, **kwargs):
        auth = self.new_plugin(**kwargs)
        auth_ref = auth.get_auth_ref(self.session)
        self.assertIsInstance(auth_ref, access.AccessInfoV2)
        self.assertEqual(
            self.TEST_URL + 'v2.0/tokens', self.requests_mock.last_request.url
        )
        self.assertIsInstance(auth._plugin, self.V2_PLUGIN_CLASS)
        return auth

    def assertDiscoveryFailure(self, **kwargs):
        plugin = self.new_plugin(**kwargs)
        self.assertRaises(
            exceptions.DiscoveryFailure, plugin.get_auth_ref, self.session
        )

    def test_create_v3_if_domain_params(self):
        self.stub_discovery()

        self.assertCreateV3(domain_id=uuid.uuid4().hex)
        self.assertCreateV3(domain_name=uuid.uuid4().hex)
        self.assertCreateV3(
            project_name=uuid.uuid4().hex, project_domain_name=uuid.uuid4().hex
        )
        self.assertCreateV3(
            project_name=uuid.uuid4().hex, project_domain_id=uuid.uuid4().hex
        )

    def test_create_v2_if_no_domain_params(self):
        self.stub_discovery()
        self.assertCreateV2()
        self.assertCreateV2(project_id=uuid.uuid4().hex)
        self.assertCreateV2(project_name=uuid.uuid4().hex)
        self.assertCreateV2(tenant_id=uuid.uuid4().hex)
        self.assertCreateV2(tenant_name=uuid.uuid4().hex)

    def test_create_plugin_no_reauthenticate(self):
        self.stub_discovery()
        self.assertCreateV2(reauthenticate=False)
        self.assertCreateV3(domain_id=uuid.uuid4().hex, reauthenticate=False)

    def test_v3_params_v2_url(self):
        self.stub_discovery(v3=False)
        self.assertDiscoveryFailure(domain_name=uuid.uuid4().hex)

    def test_v2_params_v3_url(self):
        self.stub_discovery(v2=False)
        self.assertCreateV3()

    def test_no_urls(self):
        self.stub_discovery(v2=False, v3=False)
        self.assertDiscoveryFailure()

    def test_path_based_url_v2(self):
        self.stub_url('GET', ['v2.0'], status_code=403)
        self.assertCreateV2(auth_url=self.TEST_URL + 'v2.0')

    def test_path_based_url_v3(self):
        self.stub_url('GET', ['v3'], status_code=403)
        self.assertCreateV3(auth_url=self.TEST_URL + 'v3')

    def test_disc_error_for_failure(self):
        self.stub_url('GET', [], status_code=403)
        self.assertDiscoveryFailure()
        self.assertIn(self.TEST_URL, self.logger.output)

    def test_v3_plugin_from_failure(self):
        url = self.TEST_URL + 'v3'
        self.stub_url('GET', [], base_url=url, status_code=403)
        self.assertCreateV3(auth_url=url)

    def test_unknown_discovery_version(self):
        # make a v4 entry that's mostly the same as a v3
        self.stub_discovery(v2=False, v3_id='v4.0')
        self.assertDiscoveryFailure()

    def test_default_domain_id_with_v3(self, **kwargs):
        self.stub_discovery()
        project_name = uuid.uuid4().hex
        default_domain_id = kwargs.setdefault(
            'default_domain_id', uuid.uuid4().hex
        )

        p = self.assertCreateV3(project_name=project_name, **kwargs)

        self.assertEqual(default_domain_id, p._plugin.project_domain_id)
        self.assertEqual(project_name, p._plugin.project_name)

        return p

    def test_default_domain_id_no_v3(self):
        self.stub_discovery(v3=False)
        project_name = uuid.uuid4().hex
        default_domain_id = uuid.uuid4().hex

        p = self.assertCreateV2(
            project_name=project_name, default_domain_id=default_domain_id
        )

        self.assertEqual(project_name, p._plugin.tenant_name)

    def test_default_domain_name_with_v3(self, **kwargs):
        self.stub_discovery()
        project_name = uuid.uuid4().hex
        default_domain_name = kwargs.setdefault(
            'default_domain_name', uuid.uuid4().hex
        )

        p = self.assertCreateV3(project_name=project_name, **kwargs)

        self.assertEqual(default_domain_name, p._plugin.project_domain_name)
        self.assertEqual(project_name, p._plugin.project_name)

        return p

    def test_default_domain_name_no_v3(self):
        self.stub_discovery(v3=False)
        project_name = uuid.uuid4().hex
        default_domain_name = uuid.uuid4().hex

        p = self.assertCreateV2(
            project_name=project_name, default_domain_name=default_domain_name
        )

        self.assertEqual(project_name, p._plugin.tenant_name)
